import os
import unittest
from math import inf
from xml.etree.ElementTree import parse, ParseError

from sierralocal.hivdb import HIVdb

class TestHIVDb(unittest.TestCase):
    def setUp(self):
        self.algorithm = HIVdb()
        self.root = parse(self.algorithm.xml_filename)       

    # These were written for HIVDB version 9.4, texts may change
    def testSetHivdbXml(self):
        exp_xml_filename = os.path.abspath(r'sierralocal/data/HIVDB_9.4.xml')
        res_xml_filename = os.path.abspath(self.algorithm.xml_filename)
        self.assertEqual(exp_xml_filename, res_xml_filename)

        with self.assertRaises(FileNotFoundError) as err:
            invalid_path = r'does/not/exist.txt'
            algorithm = HIVdb(asi2=invalid_path)
        
        with self.assertRaises(ParseError):
            invalid_xml = r'sierralocal/data/RT-comments.csv'
            algorithm = HIVdb(asi2=invalid_xml)

    def testSetApobecJson(self):
        exp_json_filename = os.path.abspath(r'sierralocal/data/apobec_drms.json')
        res_json_filename = os.path.abspath(self.algorithm.json_filename)
        self.assertEqual(exp_json_filename, res_json_filename)

        # Setting params
        json_path = r'sierralocal\data\apobec_drms.json'

        algorithm = HIVdb(apobec=json_path)
        exp_json_filename = os.path.abspath(json_path)
        res_json_filename = os.path.abspath(algorithm.json_filename)
        self.assertEqual(exp_json_filename, res_json_filename)

    def testParseDefinitions(self):
        # exp_def_dict =    
        res_def_dict = self.algorithm.parse_definitions(self.root)

        # self.assertDictEqual(exp_def_dict, res_def_dict)
        self.assertIn('gene', res_def_dict)
        self.assertIn('drugclass', res_def_dict)
        self.assertIn('globalrange', res_def_dict)
        self.assertIn('level', res_def_dict)
        self.assertIn('comment', res_def_dict)

    def testParseGlobalrange(self):
        exp_grange_dict = {1: {'min': -inf, 'max': 9},
                    2: {'min': 10, 'max': 14},
                    3: {'min': 15, 'max': 29},
                    4: {'min': 30, 'max': 59},
                    5: {'min': 60, 'max': inf}}

        # Setting params
        globalrange_element = self.root.find('.//GLOBALRANGE')
        scorerange = globalrange_element.text.split(',')
        res_grange_dict = self.algorithm.parse_globalrange({}, scorerange)

        self.assertEqual(res_grange_dict, exp_grange_dict)

    def testParseDrugs(self):
        self.definitions = self.algorithm.parse_definitions(self.root)

        exp_drugs = \
                {'3TC': ([[{'group': [(65, 'N')], 'value': 15},
                           {'group': [(65, 'R')], 'value': 30}],
                          {'group': [(67, 'd')], 'value': 15},
                          {'group': [(68, 'd')], 'value': 15},
                          [{'group': [(69, 'i')], 'value': 30},
                           {'group': [(69, 'd')], 'value': 15}],
                          [{'group': [(70, 'E')], 'value': 10},
                           {'group': [(70, 'G')], 'value': 10},
                           {'group': [(70, 'N')], 'value': 10},
                           {'group': [(70, 'Q')], 'value': 10},
                           {'group': [(70, 'S')], 'value': 10},
                           {'group': [(70, 'T')], 'value': 10},
                           {'group': [(70, 'd')], 'value': 10}],
                          {'group': [(75, 'I')], 'value': 5},
                          {'group': [(77, 'L')], 'value': 5},
                          {'group': [(116, 'Y')], 'value': 5},
                          [{'group': [(151, 'L')], 'value': 10},
                           {'group': [(151, 'M')], 'value': 15}],
                          [{'group': [(184, 'I')], 'value': 60},
                           {'group': [(184, 'V')], 'value': 60}],
                          {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                          {'group': [(41, 'L'), (215, 'FY')], 'value': 5},
                          {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 15},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'ABC': ([{'group': [(41, 'L')], 'value': 5},
                          {'group': [(62, 'V')], 'value': 5},
                          [{'group': [(65, 'E')], 'value': 10},
                           {'group': [(65, 'N')], 'value': 30},
                           {'group': [(65, 'R')], 'value': 45}],
                          [{'group': [(67, 'E')], 'value': 5},
                           {'group': [(67, 'G')], 'value': 5},
                           {'group': [(67, 'H')], 'value': 5},
                           {'group': [(67, 'N')], 'value': 5},
                           {'group': [(67, 'S')], 'value': 5},
                           {'group': [(67, 'T')], 'value': 5},
                           {'group': [(67, 'd')], 'value': 30}],
                          {'group': [(68, 'd')], 'value': 15},
                          [{'group': [(69, 'G')], 'value': 10},
                           {'group': [(69, 'i')], 'value': 60},
                           {'group': [(69, 'd')], 'value': 15}],
                          [{'group': [(70, 'E')], 'value': 15},
                           {'group': [(70, 'G')], 'value': 15},
                           {'group': [(70, 'N')], 'value': 15},
                           {'group': [(70, 'Q')], 'value': 15},
                           {'group': [(70, 'R')], 'value': 5},
                           {'group': [(70, 'S')], 'value': 15},
                           {'group': [(70, 'T')], 'value': 15},
                           {'group': [(70, 'd')], 'value': 15}],
                          [{'group': [(74, 'I')], 'value': 15},
                           {'group': [(74, 'V')], 'value': 30}],
                          {'group': [(75, 'I')], 'value': 5},
                          {'group': [(77, 'L')], 'value': 5},
                          {'group': [(115, 'F')], 'value': 30},
                          {'group': [(116, 'Y')], 'value': 5},
                          [{'group': [(151, 'L')], 'value': 30},
                           {'group': [(151, 'M')], 'value': 60}],
                          [{'group': [(184, 'I')], 'value': 15},
                           {'group': [(184, 'V')], 'value': 15}],
                          {'group': [(210, 'W')], 'value': 5},
                          [{'group': [(215, 'F')], 'value': 10},
                           {'group': [(215, 'I')], 'value': 5},
                           {'group': [(215, 'V')], 'value': 5},
                           {'group': [(215, 'Y')], 'value': 10}],
                          [{'group': [(219, 'E')], 'value': 5},
                           {'group': [(219, 'N')], 'value': 5},
                           {'group': [(219, 'Q')], 'value': 5},
                           {'group': [(219, 'R')], 'value': 5},
                           {'group': [(219, 'W')], 'value': 5}],
                          {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                          {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                          {'group': [(74, 'V'), (184, 'IV')], 'value': 15},
                          {'group': [(115, 'F'), (184, 'IV')], 'value': 15},
                          {'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                          {'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                          {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                          {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 10},
                          {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 10},
                          {'group': [(41, 'L'), (184, 'VI'), (215, 'FY')], 'value': 10},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (184, 'VI'), (219, 'ENQRW')],
                           'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'ATV/r': ([{'group': [(20, 'T')], 'value': 5},
                            [{'group': [(24, 'F')], 'value': 5},
                             {'group': [(24, 'I')], 'value': 10},
                             {'group': [(24, 'M')], 'value': 5}],
                            {'group': [(32, 'I')], 'value': 15},
                            {'group': [(33, 'F')], 'value': 5},
                            [{'group': [(46, 'I')], 'value': 10},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 10}],
                            {'group': [(47, 'V')], 'value': 10},
                            [{'group': [(48, 'A')], 'value': 10},
                             {'group': [(48, 'L')], 'value': 10},
                             {'group': [(48, 'M')], 'value': 30},
                             {'group': [(48, 'Q')], 'value': 10},
                             {'group': [(48, 'S')], 'value': 10},
                             {'group': [(48, 'T')], 'value': 10},
                             {'group': [(48, 'V')], 'value': 30}],
                            {'group': [(50, 'L')], 'value': 60},
                            {'group': [(53, 'L')], 'value': 10},
                            [{'group': [(54, 'A')], 'value': 15},
                             {'group': [(54, 'L')], 'value': 15},
                             {'group': [(54, 'M')], 'value': 15},
                             {'group': [(54, 'S')], 'value': 15},
                             {'group': [(54, 'T')], 'value': 15},
                             {'group': [(54, 'V')], 'value': 15}],
                            [{'group': [(73, 'A')], 'value': 10},
                             {'group': [(73, 'C')], 'value': 10},
                             {'group': [(73, 'D')], 'value': 5},
                             {'group': [(73, 'S')], 'value': 10},
                             {'group': [(73, 'T')], 'value': 10},
                             {'group': [(73, 'V')], 'value': 5}],
                            {'group': [(74, 'P')], 'value': 10},
                            [{'group': [(82, 'A')], 'value': 15},
                             {'group': [(82, 'C')], 'value': 15},
                             {'group': [(82, 'F')], 'value': 15},
                             {'group': [(82, 'L')], 'value': 10},
                             {'group': [(82, 'M')], 'value': 10},
                             {'group': [(82, 'S')], 'value': 30},
                             {'group': [(82, 'T')], 'value': 30}],
                            {'group': [(83, 'D')], 'value': 10},
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 60},
                             {'group': [(84, 'V')], 'value': 60}],
                            [{'group': [(88, 'D')], 'value': 10},
                             {'group': [(88, 'G')], 'value': 15},
                             {'group': [(88, 'S')], 'value': 60},
                             {'group': [(88, 'T')], 'value': 15}],
                            {'group': [(90, 'M')], 'value': 25},
                            {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                            {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                            {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                            {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                            {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                            {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                            {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'AZT': ([{'group': [(41, 'L')], 'value': 15},
                          {'group': [(62, 'V')], 'value': 5},
                          {'group': [(65, 'R')], 'value': -10},
                          [{'group': [(67, 'E')], 'value': 15},
                           {'group': [(67, 'G')], 'value': 15},
                           {'group': [(67, 'H')], 'value': 15},
                           {'group': [(67, 'N')], 'value': 15},
                           {'group': [(67, 'S')], 'value': 15},
                           {'group': [(67, 'T')], 'value': 15},
                           {'group': [(67, 'd')], 'value': 30}],
                          [{'group': [(69, 'G')], 'value': 5},
                           {'group': [(69, 'i')], 'value': 60}],
                          {'group': [(70, 'R')], 'value': 30},
                          [{'group': [(75, 'I')], 'value': 5},
                           {'group': [(75, 'M')], 'value': 10},
                           {'group': [(75, 'A')], 'value': 10},
                           {'group': [(75, 'S')], 'value': 10},
                           {'group': [(75, 'T')], 'value': 10}],
                          {'group': [(77, 'L')], 'value': 10},
                          {'group': [(116, 'Y')], 'value': 10},
                          [{'group': [(151, 'L')], 'value': 30},
                           {'group': [(151, 'M')], 'value': 60}],
                          [{'group': [(184, 'I')], 'value': -10},
                           {'group': [(184, 'V')], 'value': -10}],
                          {'group': [(210, 'W')], 'value': 15},
                          [{'group': [(215, 'A')], 'value': 10},
                           {'group': [(215, 'C')], 'value': 10},
                           {'group': [(215, 'D')], 'value': 10},
                           {'group': [(215, 'E')], 'value': 10},
                           {'group': [(215, 'F')], 'value': 60},
                           {'group': [(215, 'I')], 'value': 20},
                           {'group': [(215, 'L')], 'value': 10},
                           {'group': [(215, 'N')], 'value': 10},
                           {'group': [(215, 'S')], 'value': 10},
                           {'group': [(215, 'V')], 'value': 20},
                           {'group': [(215, 'Y')], 'value': 60}],
                          [{'group': [(219, 'E')], 'value': 10},
                           {'group': [(219, 'N')], 'value': 10},
                           {'group': [(219, 'Q')], 'value': 10},
                           {'group': [(219, 'R')], 'value': 10},
                           {'group': [(219, 'W')], 'value': 10}],
                          {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                          {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 10},
                          [{'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                           {'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 10}],
                          [{'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                           {'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 10}],
                          {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                          {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 15}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'BIC': ([{'group': [(51, 'Y')], 'value': 10},
                          [{'group': [(66, 'I')], 'value': 5},
                           {'group': [(66, 'K')], 'value': 15}],
                          {'group': [(92, 'Q')], 'value': 10},
                          {'group': [(118, 'R')], 'value': 30},
                          [{'group': [(121, 'C')], 'value': 15},
                           {'group': [(121, 'Y')], 'value': 10}],
                          [{'group': [(138, 'A')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'T')], 'value': 10}],
                          [{'group': [(140, 'A')], 'value': 10},
                           {'group': [(140, 'C')], 'value': 10},
                           {'group': [(140, 'R')], 'value': 10},
                           {'group': [(140, 'S')], 'value': 10}],
                          [{'group': [(143, 'A')], 'value': 5},
                           {'group': [(143, 'C')], 'value': 5},
                           {'group': [(143, 'G')], 'value': 5},
                           {'group': [(143, 'H')], 'value': 5},
                           {'group': [(143, 'K')], 'value': 5},
                           {'group': [(143, 'R')], 'value': 5},
                           {'group': [(143, 'S')], 'value': 5}],
                          {'group': [(147, 'G')], 'value': 10},
                          [{'group': [(148, 'H')], 'value': 25},
                           {'group': [(148, 'K')], 'value': 30},
                           {'group': [(148, 'R')], 'value': 25}],
                          {'group': [(151, 'L')], 'value': 15},
                          [{'group': [(153, 'F')], 'value': 15},
                           {'group': [(153, 'Y')], 'value': 15}],
                          {'group': [(155, 'H')], 'value': 10},
                          {'group': [(230, 'R')], 'value': 10},
                          {'group': [(263, 'K')], 'value': 30},
                          {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                          {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 10},
                          {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 10},
                          {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 10},
                          {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')], 'value': 10},
                          {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                          {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                          {'group': [(147, 'G'), (148, 'HKR')], 'value': 15},
                          {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                          {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                          {'group': [(148, 'HKR'), (163, 'KR')], 'value': 5},
                          {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                          {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                          {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                          {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                          {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                          {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                          {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                          {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                          {'group': [(97, 'A'), (148, 'HKR')], 'value': 15}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'CAB': ([{'group': [(51, 'Y')], 'value': 15},
                          [{'group': [(66, 'I')], 'value': 10},
                           {'group': [(66, 'K')], 'value': 20}],
                          [{'group': [(92, 'G')], 'value': 10},
                           {'group': [(92, 'Q')], 'value': 15},
                           {'group': [(92, 'V')], 'value': 10}],
                          {'group': [(118, 'R')], 'value': 60},
                          [{'group': [(121, 'C')], 'value': 60},
                           {'group': [(121, 'Y')], 'value': 15}],
                          [{'group': [(138, 'A')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'T')], 'value': 10}],
                          [{'group': [(140, 'A')], 'value': 10},
                           {'group': [(140, 'C')], 'value': 10},
                           {'group': [(140, 'R')], 'value': 60},
                           {'group': [(140, 'S')], 'value': 10}],
                          [{'group': [(143, 'A')], 'value': 10},
                           {'group': [(143, 'C')], 'value': 10},
                           {'group': [(143, 'G')], 'value': 10},
                           {'group': [(143, 'H')], 'value': 10},
                           {'group': [(143, 'K')], 'value': 10},
                           {'group': [(143, 'R')], 'value': 10},
                           {'group': [(143, 'S')], 'value': 10}],
                          {'group': [(146, 'P')], 'value': 30},
                          {'group': [(147, 'G')], 'value': 10},
                          [{'group': [(148, 'H')], 'value': 30},
                           {'group': [(148, 'K')], 'value': 50},
                           {'group': [(148, 'N')], 'value': 15},
                           {'group': [(148, 'R')], 'value': 40}],
                          {'group': [(151, 'L')], 'value': 15},
                          [{'group': [(153, 'F')], 'value': 25},
                           {'group': [(153, 'Y')], 'value': 25}],
                          [{'group': [(155, 'H')], 'value': 25},
                           {'group': [(155, 'S')], 'value': 15},
                           {'group': [(155, 'T')], 'value': 15}],
                          {'group': [(230, 'R')], 'value': 20},
                          {'group': [(263, 'K')], 'value': 30},
                          {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                          {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                          {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 20},
                          {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 20},
                          {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')], 'value': 10},
                          {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                          {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                          {'group': [(147, 'G'), (148, 'HKR')], 'value': 20},
                          {'group': [(147, 'G'), (155, 'H')], 'value': 25},
                          {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                          {'group': [(148, 'HKR'), (163, 'KR')], 'value': 20},
                          {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                          {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                          {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                          {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                          {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 10},
                          {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                          {'group': [(92, 'Q'), (155, 'H')], 'value': 20},
                          {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                          {'group': [(97, 'A'), (143, 'ACGHRS')], 'value': 5},
                          {'group': [(97, 'A'), (148, 'HKR')], 'value': 20}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'D4T': ([{'group': [(41, 'L')], 'value': 15},
                          {'group': [(62, 'V')], 'value': 5},
                          [{'group': [(65, 'E')], 'value': 10},
                           {'group': [(65, 'N')], 'value': 30},
                           {'group': [(65, 'R')], 'value': 60}],
                          [{'group': [(67, 'E')], 'value': 10},
                           {'group': [(67, 'G')], 'value': 10},
                           {'group': [(67, 'H')], 'value': 10},
                           {'group': [(67, 'N')], 'value': 15},
                           {'group': [(67, 'S')], 'value': 10},
                           {'group': [(67, 'T')], 'value': 10},
                           {'group': [(67, 'd')], 'value': 30}],
                          {'group': [(68, 'd')], 'value': 30},
                          [{'group': [(69, 'G')], 'value': 10},
                           {'group': [(69, 'i')], 'value': 60},
                           {'group': [(69, 'd')], 'value': 30},
                           {'group': [(69, 'D')], 'value': 10}],
                          [{'group': [(70, 'E')], 'value': 15},
                           {'group': [(70, 'G')], 'value': 15},
                           {'group': [(70, 'N')], 'value': 15},
                           {'group': [(70, 'Q')], 'value': 15},
                           {'group': [(70, 'R')], 'value': 15},
                           {'group': [(70, 'S')], 'value': 15},
                           {'group': [(70, 'T')], 'value': 15},
                           {'group': [(70, 'd')], 'value': 30}],
                          [{'group': [(75, 'I')], 'value': 5},
                           {'group': [(75, 'M')], 'value': 30},
                           {'group': [(75, 'A')], 'value': 30},
                           {'group': [(75, 'S')], 'value': 30},
                           {'group': [(75, 'T')], 'value': 60}],
                          {'group': [(77, 'L')], 'value': 10},
                          {'group': [(116, 'Y')], 'value': 10},
                          [{'group': [(151, 'L')], 'value': 30},
                           {'group': [(151, 'M')], 'value': 60}],
                          [{'group': [(184, 'I')], 'value': -10},
                           {'group': [(184, 'V')], 'value': -10}],
                          {'group': [(210, 'W')], 'value': 15},
                          [{'group': [(215, 'A')], 'value': 20},
                           {'group': [(215, 'C')], 'value': 20},
                           {'group': [(215, 'D')], 'value': 20},
                           {'group': [(215, 'E')], 'value': 20},
                           {'group': [(215, 'F')], 'value': 40},
                           {'group': [(215, 'I')], 'value': 20},
                           {'group': [(215, 'L')], 'value': 20},
                           {'group': [(215, 'N')], 'value': 20},
                           {'group': [(215, 'S')], 'value': 20},
                           {'group': [(215, 'V')], 'value': 20},
                           {'group': [(215, 'Y')], 'value': 40}],
                          [{'group': [(219, 'E')], 'value': 10},
                           {'group': [(219, 'N')], 'value': 10},
                           {'group': [(219, 'Q')], 'value': 10},
                           {'group': [(219, 'R')], 'value': 10},
                           {'group': [(219, 'W')], 'value': 10}],
                          {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                          {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (67, 'EGN'), (215, 'FY')], 'value': 5},
                          {'group': [(67, 'EGN'), (215, 'FY'), (219, 'ENQR')], 'value': 5},
                          {'group': [(67, 'EGN'), (70, 'R'), (219, 'ENQR')], 'value': 10},
                          {'group': [(70, 'R'), (215, 'FY')], 'value': 5},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'M')], 'value': 10},
                          [{'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 5},
                           {'group': [(210, 'W'), (215, 'FY')], 'value': 10}],
                          [{'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 5},
                           {'group': [(41, 'L'), (215, 'FY')], 'value': 10}],
                          {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                          {'group': [(70, 'EGNQST'), (184, 'IV')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'DDI': ([{'group': [(41, 'L')], 'value': 10},
                          {'group': [(62, 'V')], 'value': 5},
                          [{'group': [(65, 'E')], 'value': 10},
                           {'group': [(65, 'N')], 'value': 30},
                           {'group': [(65, 'R')], 'value': 60}],
                          [{'group': [(67, 'E')], 'value': 5},
                           {'group': [(67, 'G')], 'value': 5},
                           {'group': [(67, 'H')], 'value': 5},
                           {'group': [(67, 'N')], 'value': 5},
                           {'group': [(67, 'S')], 'value': 5},
                           {'group': [(67, 'T')], 'value': 5},
                           {'group': [(67, 'd')], 'value': 30}],
                          {'group': [(68, 'd')], 'value': 30},
                          [{'group': [(69, 'G')], 'value': 10},
                           {'group': [(69, 'i')], 'value': 60},
                           {'group': [(69, 'd')], 'value': 30},
                           {'group': [(69, 'D')], 'value': 30}],
                          [{'group': [(70, 'E')], 'value': 15},
                           {'group': [(70, 'G')], 'value': 15},
                           {'group': [(70, 'N')], 'value': 15},
                           {'group': [(70, 'Q')], 'value': 15},
                           {'group': [(70, 'R')], 'value': 10},
                           {'group': [(70, 'S')], 'value': 15},
                           {'group': [(70, 'T')], 'value': 15},
                           {'group': [(70, 'd')], 'value': 30}],
                          [{'group': [(74, 'I')], 'value': 60},
                           {'group': [(74, 'V')], 'value': 60}],
                          [{'group': [(75, 'I')], 'value': 5},
                           {'group': [(75, 'M')], 'value': 15},
                           {'group': [(75, 'A')], 'value': 15},
                           {'group': [(75, 'S')], 'value': 15},
                           {'group': [(75, 'T')], 'value': 30}],
                          {'group': [(77, 'L')], 'value': 10},
                          {'group': [(116, 'Y')], 'value': 10},
                          [{'group': [(151, 'L')], 'value': 30},
                           {'group': [(151, 'M')], 'value': 60}],
                          [{'group': [(184, 'I')], 'value': 10},
                           {'group': [(184, 'V')], 'value': 10}],
                          {'group': [(210, 'W')], 'value': 10},
                          [{'group': [(215, 'A')], 'value': 10},
                           {'group': [(215, 'C')], 'value': 10},
                           {'group': [(215, 'D')], 'value': 10},
                           {'group': [(215, 'E')], 'value': 10},
                           {'group': [(215, 'F')], 'value': 15},
                           {'group': [(215, 'I')], 'value': 10},
                           {'group': [(215, 'L')], 'value': 10},
                           {'group': [(215, 'N')], 'value': 10},
                           {'group': [(215, 'S')], 'value': 10},
                           {'group': [(215, 'V')], 'value': 10},
                           {'group': [(215, 'Y')], 'value': 15}],
                          [{'group': [(219, 'E')], 'value': 5},
                           {'group': [(219, 'N')], 'value': 5},
                           {'group': [(219, 'Q')], 'value': 5},
                           {'group': [(219, 'R')], 'value': 5},
                           {'group': [(219, 'W')], 'value': 5}],
                          {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                          {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (67, 'EGN'), (215, 'FY')], 'value': 5},
                          {'group': [(67, 'EGN'), (215, 'FY'), (219, 'ENQR')], 'value': 5},
                          {'group': [(67, 'EGN'), (70, 'R'), (219, 'ENQR')], 'value': 10},
                          {'group': [(70, 'R'), (215, 'FY')], 'value': 5},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'M')], 'value': 10},
                          [{'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 5},
                           {'group': [(210, 'W'), (215, 'FY')], 'value': 10}],
                          [{'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 5},
                           {'group': [(41, 'L'), (215, 'FY')], 'value': 10}]],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'DOR': ([{'group': [(98, 'G')], 'value': 15},
                          [{'group': [(100, 'I')], 'value': 15},
                           {'group': [(100, 'V')], 'value': 10}],
                          [{'group': [(101, 'E')], 'value': 15},
                           {'group': [(101, 'P')], 'value': 10}],
                          [{'group': [(106, 'A')], 'value': 60},
                           {'group': [(106, 'I')], 'value': 10},
                           {'group': [(106, 'M')], 'value': 30}],
                          {'group': [(108, 'I')], 'value': 10},
                          {'group': [(138, 'K')], 'value': 5},
                          {'group': [(179, 'F')], 'value': 10},
                          [{'group': [(181, 'C')], 'value': 10},
                           {'group': [(181, 'I')], 'value': 20},
                           {'group': [(181, 'V')], 'value': 20}],
                          [{'group': [(188, 'C')], 'value': 5},
                           {'group': [(188, 'F')], 'value': 30},
                           {'group': [(188, 'H')], 'value': 5},
                           {'group': [(188, 'L')], 'value': 60}],
                          [{'group': [(190, 'C')], 'value': 10},
                           {'group': [(190, 'E')], 'value': 60},
                           {'group': [(190, 'Q')], 'value': 30},
                           {'group': [(190, 'S')], 'value': 20},
                           {'group': [(190, 'T')], 'value': 10},
                           {'group': [(190, 'V')], 'value': 10}],
                          {'group': [(221, 'Y')], 'value': 10},
                          {'group': [(225, 'H')], 'value': 20},
                          [{'group': [(227, 'C')], 'value': 60},
                           {'group': [(227, 'I')], 'value': 60},
                           {'group': [(227, 'L')], 'value': 60},
                           {'group': [(227, 'V')], 'value': 60}],
                          [{'group': [(230, 'I')], 'value': 15},
                           {'group': [(230, 'L')], 'value': 60}],
                          {'group': [(234, 'I')], 'value': 45},
                          {'group': [(236, 'L')], 'value': 10},
                          {'group': [(318, 'F')], 'value': 60},
                          {'group': [(100, 'I'), (103, 'N')], 'value': 15},
                          {'group': [(100, 'I'), (103, 'N'), (225, 'H')], 'value': -10},
                          {'group': [(103, 'N'), (181, 'C')], 'value': 5},
                          {'group': [(103, 'N'), (225, 'H')], 'value': 10},
                          {'group': [(106, 'I'), (181, 'C')], 'value': 5},
                          {'group': [(106, 'I'), (190, 'S')], 'value': 5},
                          {'group': [(108, 'I'), (181, 'C')], 'value': 5},
                          {'group': [(108, 'I'), (234, 'I')], 'value': 10},
                          {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                          {'group': [(181, 'CIV'), (221, 'Y')], 'value': 10},
                          {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                          {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                          [{'group': [(101, 'E'), (190, 'A')], 'value': 5},
                           {'group': [(101, 'E'), (190, 'S')], 'value': 5}]],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'DRV/r': ([{'group': [(10, 'F')], 'value': 5},
                            {'group': [(32, 'I')], 'value': 15},
                            {'group': [(33, 'F')], 'value': 5},
                            [{'group': [(47, 'A')], 'value': 10},
                             {'group': [(47, 'V')], 'value': 10}],
                            [{'group': [(50, 'L')], 'value': -10},
                             {'group': [(50, 'V')], 'value': 20}],
                            [{'group': [(54, 'L')], 'value': 20},
                             {'group': [(54, 'M')], 'value': 20}],
                            {'group': [(74, 'P')], 'value': 5},
                            {'group': [(76, 'V')], 'value': 20},
                            {'group': [(82, 'F')], 'value': 15},
                            [{'group': [(84, 'A')], 'value': 30},
                             {'group': [(84, 'C')], 'value': 15},
                             {'group': [(84, 'V')], 'value': 15}],
                            {'group': [(88, 'S')], 'value': -5},
                            {'group': [(89, 'V')], 'value': 5},
                            {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                            {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                            {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                            {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                            {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'LM'), (89, 'V')], 'value': 5}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'DTG': ([{'group': [(51, 'Y')], 'value': 10},
                          [{'group': [(66, 'I')], 'value': 5},
                           {'group': [(66, 'K')], 'value': 15}],
                          {'group': [(92, 'Q')], 'value': 10},
                          {'group': [(118, 'R')], 'value': 50},
                          [{'group': [(121, 'C')], 'value': 15},
                           {'group': [(121, 'Y')], 'value': 10}],
                          [{'group': [(138, 'A')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'T')], 'value': 10}],
                          [{'group': [(140, 'A')], 'value': 10},
                           {'group': [(140, 'C')], 'value': 10},
                           {'group': [(140, 'R')], 'value': 10},
                           {'group': [(140, 'S')], 'value': 10}],
                          [{'group': [(143, 'A')], 'value': 5},
                           {'group': [(143, 'C')], 'value': 5},
                           {'group': [(143, 'G')], 'value': 5},
                           {'group': [(143, 'H')], 'value': 5},
                           {'group': [(143, 'K')], 'value': 5},
                           {'group': [(143, 'R')], 'value': 5},
                           {'group': [(143, 'S')], 'value': 5}],
                          {'group': [(147, 'G')], 'value': 10},
                          [{'group': [(148, 'H')], 'value': 25},
                           {'group': [(148, 'K')], 'value': 30},
                           {'group': [(148, 'R')], 'value': 25}],
                          {'group': [(151, 'L')], 'value': 15},
                          [{'group': [(153, 'F')], 'value': 15},
                           {'group': [(153, 'Y')], 'value': 15}],
                          {'group': [(155, 'H')], 'value': 10},
                          {'group': [(230, 'R')], 'value': 20},
                          {'group': [(263, 'K')], 'value': 30},
                          {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                          {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 10},
                          {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 10},
                          {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 10},
                          {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')], 'value': 10},
                          {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                          {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                          {'group': [(147, 'G'), (148, 'HKR')], 'value': 15},
                          {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                          {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                          {'group': [(148, 'HKR'), (163, 'KR')], 'value': 5},
                          {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                          {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                          {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                          {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                          {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                          {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                          {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                          {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                          {'group': [(97, 'A'), (148, 'HKR')], 'value': 15}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'EFV': ([{'group': [(98, 'G')], 'value': 15},
                          [{'group': [(100, 'I')], 'value': 60},
                           {'group': [(100, 'V')], 'value': 30}],
                          [{'group': [(101, 'E')], 'value': 15},
                           {'group': [(101, 'H')], 'value': 10},
                           {'group': [(101, 'P')], 'value': 60}],
                          [{'group': [(103, 'H')], 'value': 60},
                           {'group': [(103, 'N')], 'value': 60},
                           {'group': [(103, 'S')], 'value': 45},
                           {'group': [(103, 'T')], 'value': 15}],
                          [{'group': [(106, 'A')], 'value': 45},
                           {'group': [(106, 'M')], 'value': 60}],
                          {'group': [(108, 'I')], 'value': 10},
                          [{'group': [(138, 'G')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'Q')], 'value': 10},
                           {'group': [(138, 'R')], 'value': 10}],
                          [{'group': [(179, 'D')], 'value': 10},
                           {'group': [(179, 'E')], 'value': 10},
                           {'group': [(179, 'F')], 'value': 10},
                           {'group': [(179, 'L')], 'value': 10}],
                          [{'group': [(181, 'C')], 'value': 30},
                           {'group': [(181, 'F')], 'value': 15},
                           {'group': [(181, 'G')], 'value': 15},
                           {'group': [(181, 'I')], 'value': 30},
                           {'group': [(181, 'S')], 'value': 15},
                           {'group': [(181, 'V')], 'value': 30}],
                          [{'group': [(188, 'C')], 'value': 60},
                           {'group': [(188, 'F')], 'value': 60},
                           {'group': [(188, 'H')], 'value': 30},
                           {'group': [(188, 'L')], 'value': 60}],
                          [{'group': [(190, 'A')], 'value': 45},
                           {'group': [(190, 'C')], 'value': 60},
                           {'group': [(190, 'E')], 'value': 60},
                           {'group': [(190, 'Q')], 'value': 60},
                           {'group': [(190, 'S')], 'value': 60},
                           {'group': [(190, 'T')], 'value': 60},
                           {'group': [(190, 'V')], 'value': 60}],
                          {'group': [(221, 'Y')], 'value': 10},
                          {'group': [(225, 'H')], 'value': 45},
                          [{'group': [(227, 'C')], 'value': 45},
                           {'group': [(227, 'I')], 'value': 10},
                           {'group': [(227, 'L')], 'value': 15},
                           {'group': [(227, 'V')], 'value': 10}],
                          [{'group': [(230, 'I')], 'value': 15},
                           {'group': [(230, 'L')], 'value': 45}],
                          [{'group': [(238, 'N')], 'value': 10},
                           {'group': [(238, 'T')], 'value': 30}],
                          {'group': [(318, 'F')], 'value': 10},
                          {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                          {'group': [(103, 'R'), (179, 'D')], 'value': 20},
                          {'group': [(106, 'A'), (227, 'CL')], 'value': 15},
                          {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                          {'group': [(98, 'G'), (227, 'CL')], 'value': 15}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'ETR': ([{'group': [(98, 'G')], 'value': 10},
                          [{'group': [(100, 'I')], 'value': 30},
                           {'group': [(100, 'V')], 'value': 10}],
                          [{'group': [(101, 'E')], 'value': 15},
                           {'group': [(101, 'H')], 'value': 10},
                           {'group': [(101, 'P')], 'value': 60}],
                          {'group': [(106, 'I')], 'value': 10},
                          [{'group': [(138, 'A')], 'value': 10},
                           {'group': [(138, 'G')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'Q')], 'value': 10},
                           {'group': [(138, 'R')], 'value': 10}],
                          [{'group': [(179, 'D')], 'value': 10},
                           {'group': [(179, 'E')], 'value': 10},
                           {'group': [(179, 'F')], 'value': 15},
                           {'group': [(179, 'L')], 'value': 10}],
                          [{'group': [(181, 'C')], 'value': 30},
                           {'group': [(181, 'F')], 'value': 15},
                           {'group': [(181, 'G')], 'value': 15},
                           {'group': [(181, 'I')], 'value': 60},
                           {'group': [(181, 'S')], 'value': 15},
                           {'group': [(181, 'V')], 'value': 60}],
                          {'group': [(188, 'L')], 'value': 10},
                          [{'group': [(190, 'A')], 'value': 10},
                           {'group': [(190, 'C')], 'value': 10},
                           {'group': [(190, 'E')], 'value': 45},
                           {'group': [(190, 'Q')], 'value': 45},
                           {'group': [(190, 'S')], 'value': 10},
                           {'group': [(190, 'T')], 'value': 10},
                           {'group': [(190, 'V')], 'value': 10}],
                          {'group': [(221, 'Y')], 'value': 10},
                          {'group': [(227, 'C')], 'value': 30},
                          [{'group': [(230, 'I')], 'value': 15},
                           {'group': [(230, 'L')], 'value': 30}],
                          {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                          {'group': [(101, 'E'), (188, 'L')], 'value': 5},
                          {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                          {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                          {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                          [{'group': [(101, 'E'), (190, 'A')], 'value': 5},
                           {'group': [(101, 'E'), (190, 'S')], 'value': 5}],
                          [{'group': [(179, 'F'), (181, 'C')], 'value': 15},
                           {'group': [(179, 'T'), (181, 'C')], 'value': 10}]],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'EVG': ([{'group': [(51, 'Y')], 'value': 15},
                          [{'group': [(66, 'A')], 'value': 60},
                           {'group': [(66, 'I')], 'value': 60},
                           {'group': [(66, 'K')], 'value': 60}],
                          [{'group': [(92, 'G')], 'value': 30},
                           {'group': [(92, 'Q')], 'value': 60},
                           {'group': [(92, 'V')], 'value': 60}],
                          {'group': [(95, 'K')], 'value': 10},
                          {'group': [(97, 'A')], 'value': 10},
                          {'group': [(118, 'R')], 'value': 60},
                          [{'group': [(121, 'C')], 'value': 60},
                           {'group': [(121, 'Y')], 'value': 60}],
                          [{'group': [(138, 'A')], 'value': 15},
                           {'group': [(138, 'K')], 'value': 15},
                           {'group': [(138, 'T')], 'value': 15}],
                          [{'group': [(140, 'A')], 'value': 30},
                           {'group': [(140, 'C')], 'value': 30},
                           {'group': [(140, 'R')], 'value': 30},
                           {'group': [(140, 'S')], 'value': 30}],
                          [{'group': [(143, 'A')], 'value': 10},
                           {'group': [(143, 'C')], 'value': 10},
                           {'group': [(143, 'G')], 'value': 10},
                           {'group': [(143, 'H')], 'value': 10},
                           {'group': [(143, 'K')], 'value': 10},
                           {'group': [(143, 'R')], 'value': 10},
                           {'group': [(143, 'S')], 'value': 10}],
                          {'group': [(145, 'S')], 'value': 60},
                          {'group': [(146, 'P')], 'value': 60},
                          {'group': [(147, 'G')], 'value': 60},
                          [{'group': [(148, 'H')], 'value': 60},
                           {'group': [(148, 'K')], 'value': 60},
                           {'group': [(148, 'N')], 'value': 15},
                           {'group': [(148, 'R')], 'value': 60}],
                          [{'group': [(151, 'A')], 'value': 30},
                           {'group': [(151, 'L')], 'value': 60}],
                          [{'group': [(153, 'F')], 'value': 25},
                           {'group': [(153, 'Y')], 'value': 25}],
                          [{'group': [(155, 'H')], 'value': 60},
                           {'group': [(155, 'S')], 'value': 30},
                           {'group': [(155, 'T')], 'value': 30}],
                          {'group': [(157, 'Q')], 'value': 10},
                          [{'group': [(163, 'K')], 'value': 15},
                           {'group': [(163, 'R')], 'value': 15}],
                          {'group': [(230, 'R')], 'value': 20},
                          {'group': [(232, 'N')], 'value': 10},
                          {'group': [(263, 'K')], 'value': 30},
                          {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                          {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                          {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')], 'value': 10},
                          {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                          {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                          {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                          {'group': [(51, 'Y'), (263, 'K')], 'value': 15},
                          {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                          {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                          {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                          {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                          {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                          {'group': [(97, 'A'), (143, 'ACGHRS')], 'value': 5}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'FPV/r': ([{'group': [(10, 'F')], 'value': 15},
                            {'group': [(20, 'T')], 'value': 5},
                            [{'group': [(24, 'F')], 'value': 5},
                             {'group': [(24, 'I')], 'value': 10},
                             {'group': [(24, 'M')], 'value': 5}],
                            {'group': [(32, 'I')], 'value': 30},
                            {'group': [(33, 'F')], 'value': 10},
                            [{'group': [(46, 'I')], 'value': 10},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 10}],
                            [{'group': [(47, 'A')], 'value': 60},
                             {'group': [(47, 'V')], 'value': 35}],
                            [{'group': [(50, 'L')], 'value': -5},
                             {'group': [(50, 'V')], 'value': 60}],
                            [{'group': [(54, 'A')], 'value': 10},
                             {'group': [(54, 'L')], 'value': 60},
                             {'group': [(54, 'M')], 'value': 60},
                             {'group': [(54, 'S')], 'value': 10},
                             {'group': [(54, 'T')], 'value': 10},
                             {'group': [(54, 'V')], 'value': 10}],
                            [{'group': [(73, 'A')], 'value': 10},
                             {'group': [(73, 'C')], 'value': 10},
                             {'group': [(73, 'D')], 'value': 5},
                             {'group': [(73, 'S')], 'value': 10},
                             {'group': [(73, 'T')], 'value': 10},
                             {'group': [(73, 'V')], 'value': 5}],
                            {'group': [(74, 'P')], 'value': 10},
                            {'group': [(76, 'V')], 'value': 60},
                            [{'group': [(82, 'A')], 'value': 15},
                             {'group': [(82, 'C')], 'value': 15},
                             {'group': [(82, 'F')], 'value': 30},
                             {'group': [(82, 'L')], 'value': 15},
                             {'group': [(82, 'M')], 'value': 15},
                             {'group': [(82, 'S')], 'value': 15},
                             {'group': [(82, 'T')], 'value': 15}],
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 60},
                             {'group': [(84, 'V')], 'value': 60}],
                            {'group': [(88, 'S')], 'value': -10},
                            {'group': [(89, 'V')], 'value': 10},
                            {'group': [(90, 'M')], 'value': 20},
                            {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                            {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                            {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                            {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                            {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                            {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                            {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                            {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                            {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                            {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'FTC': ([[{'group': [(65, 'N')], 'value': 15},
                           {'group': [(65, 'R')], 'value': 30}],
                          {'group': [(67, 'd')], 'value': 15},
                          {'group': [(68, 'd')], 'value': 15},
                          [{'group': [(69, 'i')], 'value': 30},
                           {'group': [(69, 'd')], 'value': 15}],
                          [{'group': [(70, 'E')], 'value': 10},
                           {'group': [(70, 'G')], 'value': 10},
                           {'group': [(70, 'N')], 'value': 10},
                           {'group': [(70, 'Q')], 'value': 10},
                           {'group': [(70, 'S')], 'value': 10},
                           {'group': [(70, 'T')], 'value': 10},
                           {'group': [(70, 'd')], 'value': 10}],
                          {'group': [(75, 'I')], 'value': 5},
                          {'group': [(77, 'L')], 'value': 5},
                          {'group': [(116, 'Y')], 'value': 5},
                          [{'group': [(151, 'L')], 'value': 10},
                           {'group': [(151, 'M')], 'value': 15}],
                          [{'group': [(184, 'I')], 'value': 60},
                           {'group': [(184, 'V')], 'value': 60}],
                          {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                          {'group': [(41, 'L'), (215, 'FY')], 'value': 5},
                          {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 15},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'IDV/r': ([{'group': [(10, 'F')], 'value': 10},
                            {'group': [(20, 'T')], 'value': 5},
                            [{'group': [(24, 'F')], 'value': 5},
                             {'group': [(24, 'I')], 'value': 15},
                             {'group': [(24, 'M')], 'value': 5}],
                            {'group': [(32, 'I')], 'value': 15},
                            {'group': [(33, 'F')], 'value': 5},
                            [{'group': [(46, 'I')], 'value': 10},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 10}],
                            [{'group': [(47, 'A')], 'value': 15},
                             {'group': [(47, 'V')], 'value': 15}],
                            [{'group': [(48, 'A')], 'value': 10},
                             {'group': [(48, 'L')], 'value': 10},
                             {'group': [(48, 'M')], 'value': 10},
                             {'group': [(48, 'Q')], 'value': 10},
                             {'group': [(48, 'S')], 'value': 10},
                             {'group': [(48, 'T')], 'value': 10},
                             {'group': [(48, 'V')], 'value': 10}],
                            {'group': [(50, 'L')], 'value': -5},
                            [{'group': [(54, 'A')], 'value': 15},
                             {'group': [(54, 'L')], 'value': 10},
                             {'group': [(54, 'M')], 'value': 15},
                             {'group': [(54, 'S')], 'value': 15},
                             {'group': [(54, 'T')], 'value': 15},
                             {'group': [(54, 'V')], 'value': 15}],
                            [{'group': [(73, 'A')], 'value': 15},
                             {'group': [(73, 'C')], 'value': 15},
                             {'group': [(73, 'D')], 'value': 5},
                             {'group': [(73, 'S')], 'value': 15},
                             {'group': [(73, 'T')], 'value': 15},
                             {'group': [(73, 'V')], 'value': 5}],
                            {'group': [(74, 'P')], 'value': 10},
                            {'group': [(76, 'V')], 'value': 30},
                            [{'group': [(82, 'A')], 'value': 30},
                             {'group': [(82, 'C')], 'value': 15},
                             {'group': [(82, 'F')], 'value': 30},
                             {'group': [(82, 'L')], 'value': 10},
                             {'group': [(82, 'M')], 'value': 30},
                             {'group': [(82, 'S')], 'value': 30},
                             {'group': [(82, 'T')], 'value': 30}],
                            {'group': [(83, 'D')], 'value': 10},
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 60},
                             {'group': [(84, 'V')], 'value': 60}],
                            {'group': [(88, 'S')], 'value': 15},
                            {'group': [(89, 'V')], 'value': 5},
                            {'group': [(90, 'M')], 'value': 30},
                            {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                            {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                            {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                            {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                            {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                            {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                            {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                            {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                            {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                            {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'LEN': ([{'group': [(9999, 'A')], 'value': 0}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'LPV/r': ([{'group': [(10, 'F')], 'value': 5},
                            [{'group': [(24, 'F')], 'value': 5},
                             {'group': [(24, 'I')], 'value': 10},
                             {'group': [(24, 'M')], 'value': 5}],
                            {'group': [(32, 'I')], 'value': 15},
                            {'group': [(33, 'F')], 'value': 5},
                            [{'group': [(46, 'I')], 'value': 10},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 5}],
                            [{'group': [(47, 'A')], 'value': 60},
                             {'group': [(47, 'V')], 'value': 15}],
                            [{'group': [(48, 'A')], 'value': 10},
                             {'group': [(48, 'L')], 'value': 10},
                             {'group': [(48, 'M')], 'value': 10},
                             {'group': [(48, 'Q')], 'value': 10},
                             {'group': [(48, 'S')], 'value': 10},
                             {'group': [(48, 'T')], 'value': 10},
                             {'group': [(48, 'V')], 'value': 10}],
                            [{'group': [(50, 'L')], 'value': -10},
                             {'group': [(50, 'V')], 'value': 30}],
                            [{'group': [(54, 'A')], 'value': 15},
                             {'group': [(54, 'L')], 'value': 20},
                             {'group': [(54, 'M')], 'value': 20},
                             {'group': [(54, 'S')], 'value': 15},
                             {'group': [(54, 'T')], 'value': 15},
                             {'group': [(54, 'V')], 'value': 15}],
                            [{'group': [(73, 'A')], 'value': 5},
                             {'group': [(73, 'C')], 'value': 5},
                             {'group': [(73, 'D')], 'value': 5},
                             {'group': [(73, 'S')], 'value': 5},
                             {'group': [(73, 'T')], 'value': 5},
                             {'group': [(73, 'V')], 'value': 5}],
                            {'group': [(74, 'P')], 'value': 5},
                            {'group': [(76, 'V')], 'value': 30},
                            [{'group': [(82, 'A')], 'value': 30},
                             {'group': [(82, 'C')], 'value': 15},
                             {'group': [(82, 'F')], 'value': 30},
                             {'group': [(82, 'L')], 'value': 10},
                             {'group': [(82, 'M')], 'value': 25},
                             {'group': [(82, 'S')], 'value': 30},
                             {'group': [(82, 'T')], 'value': 30}],
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 30},
                             {'group': [(84, 'V')], 'value': 30}],
                            {'group': [(90, 'M')], 'value': 15},
                            {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                            {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                            {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                            {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                            {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                            {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                            {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                            {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 5},
                            {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                            {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                            {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 5}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'NFV': ([{'group': [(10, 'F')], 'value': 15},
                          {'group': [(20, 'T')], 'value': 15},
                          {'group': [(23, 'I')], 'value': 15},
                          [{'group': [(24, 'F')], 'value': 10},
                           {'group': [(24, 'I')], 'value': 10},
                           {'group': [(24, 'M')], 'value': 10}],
                          {'group': [(30, 'N')], 'value': 60},
                          {'group': [(32, 'I')], 'value': 15},
                          {'group': [(33, 'F')], 'value': 10},
                          {'group': [(43, 'T')], 'value': 10},
                          [{'group': [(46, 'I')], 'value': 30},
                           {'group': [(46, 'L')], 'value': 20},
                           {'group': [(46, 'V')], 'value': 20}],
                          [{'group': [(47, 'A')], 'value': 30},
                           {'group': [(47, 'V')], 'value': 20}],
                          [{'group': [(48, 'A')], 'value': 30},
                           {'group': [(48, 'L')], 'value': 30},
                           {'group': [(48, 'M')], 'value': 30},
                           {'group': [(48, 'Q')], 'value': 30},
                           {'group': [(48, 'S')], 'value': 30},
                           {'group': [(48, 'T')], 'value': 30},
                           {'group': [(48, 'V')], 'value': 30}],
                          {'group': [(50, 'V')], 'value': 15},
                          {'group': [(53, 'L')], 'value': 10},
                          [{'group': [(54, 'A')], 'value': 20},
                           {'group': [(54, 'L')], 'value': 20},
                           {'group': [(54, 'M')], 'value': 20},
                           {'group': [(54, 'S')], 'value': 20},
                           {'group': [(54, 'T')], 'value': 20},
                           {'group': [(54, 'V')], 'value': 20}],
                          {'group': [(58, 'E')], 'value': 10},
                          [{'group': [(73, 'A')], 'value': 15},
                           {'group': [(73, 'C')], 'value': 15},
                           {'group': [(73, 'D')], 'value': 10},
                           {'group': [(73, 'S')], 'value': 15},
                           {'group': [(73, 'T')], 'value': 15},
                           {'group': [(73, 'V')], 'value': 10}],
                          {'group': [(74, 'P')], 'value': 20},
                          {'group': [(76, 'V')], 'value': 10},
                          [{'group': [(82, 'A')], 'value': 30},
                           {'group': [(82, 'C')], 'value': 30},
                           {'group': [(82, 'F')], 'value': 30},
                           {'group': [(82, 'L')], 'value': 10},
                           {'group': [(82, 'M')], 'value': 30},
                           {'group': [(82, 'S')], 'value': 30},
                           {'group': [(82, 'T')], 'value': 30}],
                          {'group': [(83, 'D')], 'value': 15},
                          [{'group': [(84, 'A')], 'value': 60},
                           {'group': [(84, 'C')], 'value': 60},
                           {'group': [(84, 'V')], 'value': 60}],
                          [{'group': [(88, 'D')], 'value': 60},
                           {'group': [(88, 'G')], 'value': 30},
                           {'group': [(88, 'S')], 'value': 60},
                           {'group': [(88, 'T')], 'value': 30}],
                          {'group': [(89, 'V')], 'value': 10},
                          {'group': [(90, 'M')], 'value': 60},
                          {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                          {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                          {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                          {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                          {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                          {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                          {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                          {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                          {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                          {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                          {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                          {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                          {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                          {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                          {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                          {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                          {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                          {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                          {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                          {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'NVP': ([{'group': [(98, 'G')], 'value': 30},
                          [{'group': [(100, 'I')], 'value': 60},
                           {'group': [(100, 'V')], 'value': 30}],
                          [{'group': [(101, 'E')], 'value': 30},
                           {'group': [(101, 'H')], 'value': 15},
                           {'group': [(101, 'P')], 'value': 60}],
                          [{'group': [(103, 'H')], 'value': 60},
                           {'group': [(103, 'N')], 'value': 60},
                           {'group': [(103, 'S')], 'value': 60},
                           {'group': [(103, 'T')], 'value': 60}],
                          [{'group': [(106, 'A')], 'value': 60},
                           {'group': [(106, 'I')], 'value': 10},
                           {'group': [(106, 'M')], 'value': 60}],
                          {'group': [(108, 'I')], 'value': 15},
                          [{'group': [(138, 'G')], 'value': 10},
                           {'group': [(138, 'K')], 'value': 10},
                           {'group': [(138, 'Q')], 'value': 10},
                           {'group': [(138, 'R')], 'value': 10}],
                          [{'group': [(179, 'D')], 'value': 10},
                           {'group': [(179, 'E')], 'value': 10},
                           {'group': [(179, 'F')], 'value': 15},
                           {'group': [(179, 'L')], 'value': 10}],
                          [{'group': [(181, 'C')], 'value': 60},
                           {'group': [(181, 'F')], 'value': 60},
                           {'group': [(181, 'G')], 'value': 60},
                           {'group': [(181, 'I')], 'value': 60},
                           {'group': [(181, 'S')], 'value': 60},
                           {'group': [(181, 'V')], 'value': 60}],
                          [{'group': [(188, 'C')], 'value': 60},
                           {'group': [(188, 'F')], 'value': 60},
                           {'group': [(188, 'H')], 'value': 60},
                           {'group': [(188, 'L')], 'value': 60}],
                          [{'group': [(190, 'A')], 'value': 60},
                           {'group': [(190, 'C')], 'value': 60},
                           {'group': [(190, 'E')], 'value': 60},
                           {'group': [(190, 'Q')], 'value': 60},
                           {'group': [(190, 'S')], 'value': 60},
                           {'group': [(190, 'T')], 'value': 60},
                           {'group': [(190, 'V')], 'value': 60}],
                          {'group': [(221, 'Y')], 'value': 15},
                          {'group': [(225, 'H')], 'value': 45},
                          [{'group': [(227, 'C')], 'value': 45},
                           {'group': [(227, 'I')], 'value': 30},
                           {'group': [(227, 'L')], 'value': 30},
                           {'group': [(227, 'V')], 'value': 30}],
                          [{'group': [(230, 'I')], 'value': 30},
                           {'group': [(230, 'L')], 'value': 60}],
                          [{'group': [(238, 'N')], 'value': 10},
                           {'group': [(238, 'T')], 'value': 30}],
                          {'group': [(318, 'F')], 'value': 30},
                          {'group': [(348, 'I')], 'value': 15},
                          {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                          {'group': [(103, 'R'), (179, 'D')], 'value': 20},
                          {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                          {'group': [(98, 'G'), (227, 'CL')], 'value': 15}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'RAL': ([{'group': [(51, 'Y')], 'value': 15},
                          [{'group': [(66, 'A')], 'value': 15},
                           {'group': [(66, 'I')], 'value': 15},
                           {'group': [(66, 'K')], 'value': 60}],
                          [{'group': [(92, 'G')], 'value': 15},
                           {'group': [(92, 'Q')], 'value': 30},
                           {'group': [(92, 'V')], 'value': 30}],
                          {'group': [(95, 'K')], 'value': 10},
                          {'group': [(97, 'A')], 'value': 10},
                          {'group': [(118, 'R')], 'value': 60},
                          [{'group': [(121, 'C')], 'value': 60},
                           {'group': [(121, 'Y')], 'value': 60}],
                          [{'group': [(138, 'A')], 'value': 15},
                           {'group': [(138, 'K')], 'value': 15},
                           {'group': [(138, 'T')], 'value': 15}],
                          [{'group': [(140, 'A')], 'value': 30},
                           {'group': [(140, 'C')], 'value': 30},
                           {'group': [(140, 'R')], 'value': 30},
                           {'group': [(140, 'S')], 'value': 30}],
                          [{'group': [(143, 'A')], 'value': 60},
                           {'group': [(143, 'C')], 'value': 60},
                           {'group': [(143, 'G')], 'value': 60},
                           {'group': [(143, 'H')], 'value': 60},
                           {'group': [(143, 'K')], 'value': 60},
                           {'group': [(143, 'R')], 'value': 60},
                           {'group': [(143, 'S')], 'value': 60}],
                          {'group': [(147, 'G')], 'value': 10},
                          [{'group': [(148, 'H')], 'value': 60},
                           {'group': [(148, 'K')], 'value': 60},
                           {'group': [(148, 'N')], 'value': 15},
                           {'group': [(148, 'R')], 'value': 60}],
                          [{'group': [(151, 'A')], 'value': 15},
                           {'group': [(151, 'L')], 'value': 30}],
                          [{'group': [(153, 'F')], 'value': 10},
                           {'group': [(153, 'Y')], 'value': 10}],
                          [{'group': [(155, 'H')], 'value': 60},
                           {'group': [(155, 'S')], 'value': 30},
                           {'group': [(155, 'T')], 'value': 30}],
                          {'group': [(157, 'Q')], 'value': 10},
                          [{'group': [(163, 'K')], 'value': 15},
                           {'group': [(163, 'R')], 'value': 15}],
                          {'group': [(230, 'R')], 'value': 20},
                          {'group': [(232, 'N')], 'value': 10},
                          {'group': [(263, 'K')], 'value': 25},
                          {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                          {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                          {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')], 'value': 10},
                          {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                          {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                          {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                          {'group': [(97, 'A'), (118, 'R')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'RPV': ([{'group': [(98, 'G')], 'value': 15},
                          [{'group': [(100, 'I')], 'value': 60},
                           {'group': [(100, 'V')], 'value': 15}],
                          [{'group': [(101, 'E')], 'value': 45},
                           {'group': [(101, 'H')], 'value': 10},
                           {'group': [(101, 'P')], 'value': 60}],
                          {'group': [(106, 'I')], 'value': 10},
                          [{'group': [(138, 'A')], 'value': 15},
                           {'group': [(138, 'G')], 'value': 15},
                           {'group': [(138, 'K')], 'value': 45},
                           {'group': [(138, 'Q')], 'value': 15},
                           {'group': [(138, 'R')], 'value': 15}],
                          [{'group': [(179, 'D')], 'value': 10},
                           {'group': [(179, 'E')], 'value': 10},
                           {'group': [(179, 'F')], 'value': 15},
                           {'group': [(179, 'L')], 'value': 15}],
                          [{'group': [(181, 'C')], 'value': 45},
                           {'group': [(181, 'F')], 'value': 30},
                           {'group': [(181, 'G')], 'value': 30},
                           {'group': [(181, 'I')], 'value': 60},
                           {'group': [(181, 'S')], 'value': 30},
                           {'group': [(181, 'V')], 'value': 60}],
                          [{'group': [(188, 'F')], 'value': 30},
                           {'group': [(188, 'L')], 'value': 60}],
                          [{'group': [(190, 'A')], 'value': 15},
                           {'group': [(190, 'C')], 'value': 10},
                           {'group': [(190, 'E')], 'value': 60},
                           {'group': [(190, 'Q')], 'value': 45},
                           {'group': [(190, 'S')], 'value': 15},
                           {'group': [(190, 'T')], 'value': 10},
                           {'group': [(190, 'V')], 'value': 10}],
                          {'group': [(221, 'Y')], 'value': 15},
                          {'group': [(227, 'C')], 'value': 45},
                          [{'group': [(230, 'I')], 'value': 30},
                           {'group': [(230, 'L')], 'value': 60}],
                          {'group': [(101, 'E'), (184, 'I')], 'value': 15},
                          {'group': [(103, 'R'), (179, 'D')], 'value': 15},
                          {'group': [(106, 'I'), (181, 'C')], 'value': 10},
                          {'group': [(138, 'K'), (184, 'I')], 'value': 15},
                          {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                          {'group': [(181, 'CIV'), (221, 'Y')], 'value': 10},
                          {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                          {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                          [{'group': [(179, 'F'), (181, 'C')], 'value': 15},
                           {'group': [(179, 'T'), (181, 'C')], 'value': 10}]],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'SQV/r': ([{'group': [(20, 'T')], 'value': 5},
                            [{'group': [(24, 'F')], 'value': 5},
                             {'group': [(24, 'I')], 'value': 10},
                             {'group': [(24, 'M')], 'value': 5}],
                            {'group': [(33, 'F')], 'value': 5},
                            [{'group': [(46, 'I')], 'value': 10},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 5}],
                            [{'group': [(48, 'A')], 'value': 60},
                             {'group': [(48, 'L')], 'value': 60},
                             {'group': [(48, 'M')], 'value': 60},
                             {'group': [(48, 'Q')], 'value': 60},
                             {'group': [(48, 'S')], 'value': 60},
                             {'group': [(48, 'T')], 'value': 60},
                             {'group': [(48, 'V')], 'value': 60}],
                            [{'group': [(50, 'L')], 'value': -5},
                             {'group': [(50, 'V')], 'value': 15}],
                            {'group': [(53, 'L')], 'value': 15},
                            [{'group': [(54, 'A')], 'value': 15},
                             {'group': [(54, 'L')], 'value': 15},
                             {'group': [(54, 'M')], 'value': 15},
                             {'group': [(54, 'S')], 'value': 15},
                             {'group': [(54, 'T')], 'value': 15},
                             {'group': [(54, 'V')], 'value': 15}],
                            [{'group': [(73, 'A')], 'value': 15},
                             {'group': [(73, 'C')], 'value': 15},
                             {'group': [(73, 'D')], 'value': 10},
                             {'group': [(73, 'S')], 'value': 15},
                             {'group': [(73, 'T')], 'value': 15},
                             {'group': [(73, 'V')], 'value': 10}],
                            {'group': [(74, 'P')], 'value': 10},
                            [{'group': [(82, 'A')], 'value': 15},
                             {'group': [(82, 'C')], 'value': 15},
                             {'group': [(82, 'F')], 'value': 10},
                             {'group': [(82, 'L')], 'value': 10},
                             {'group': [(82, 'M')], 'value': 15},
                             {'group': [(82, 'S')], 'value': 15},
                             {'group': [(82, 'T')], 'value': 15}],
                            {'group': [(83, 'D')], 'value': 10},
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 60},
                             {'group': [(84, 'V')], 'value': 60}],
                            [{'group': [(88, 'D')], 'value': 10},
                             {'group': [(88, 'S')], 'value': 15}],
                            {'group': [(90, 'M')], 'value': 45},
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                            {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(46, 'ILV'), (90, 'M')], 'value': 5},
                            {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                            {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                            {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                            {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'TDF': ([{'group': [(41, 'L')], 'value': 5},
                          {'group': [(62, 'V')], 'value': 5},
                          [{'group': [(65, 'E')], 'value': 10},
                           {'group': [(65, 'N')], 'value': 45},
                           {'group': [(65, 'R')], 'value': 50}],
                          [{'group': [(67, 'E')], 'value': 5},
                           {'group': [(67, 'G')], 'value': 5},
                           {'group': [(67, 'H')], 'value': 5},
                           {'group': [(67, 'N')], 'value': 5},
                           {'group': [(67, 'S')], 'value': 5},
                           {'group': [(67, 'T')], 'value': 5},
                           {'group': [(67, 'd')], 'value': 30}],
                          {'group': [(68, 'd')], 'value': 15},
                          [{'group': [(69, 'G')], 'value': 5},
                           {'group': [(69, 'i')], 'value': 60},
                           {'group': [(69, 'd')], 'value': 15}],
                          [{'group': [(70, 'E')], 'value': 15},
                           {'group': [(70, 'G')], 'value': 15},
                           {'group': [(70, 'N')], 'value': 15},
                           {'group': [(70, 'Q')], 'value': 15},
                           {'group': [(70, 'R')], 'value': 5},
                           {'group': [(70, 'S')], 'value': 15},
                           {'group': [(70, 'T')], 'value': 15},
                           {'group': [(70, 'd')], 'value': 15}],
                          {'group': [(74, 'I')], 'value': 5},
                          {'group': [(75, 'I')], 'value': 5},
                          {'group': [(77, 'L')], 'value': 5},
                          {'group': [(115, 'F')], 'value': 15},
                          {'group': [(116, 'Y')], 'value': 5},
                          [{'group': [(151, 'L')], 'value': 10},
                           {'group': [(151, 'M')], 'value': 15}],
                          [{'group': [(184, 'I')], 'value': -10},
                           {'group': [(184, 'V')], 'value': -10}],
                          {'group': [(210, 'W')], 'value': 5},
                          [{'group': [(215, 'F')], 'value': 10},
                           {'group': [(215, 'I')], 'value': 5},
                           {'group': [(215, 'V')], 'value': 5},
                           {'group': [(215, 'Y')], 'value': 10}],
                          [{'group': [(219, 'E')], 'value': 5},
                           {'group': [(219, 'N')], 'value': 5},
                           {'group': [(219, 'Q')], 'value': 5},
                           {'group': [(219, 'R')], 'value': 5},
                           {'group': [(219, 'W')], 'value': 5}],
                          {'group': [(151, 'M'), (184, 'IV')], 'value': 15},
                          {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                          {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                          {'group': [(115, 'F'), (184, 'IV')], 'value': 5},
                          {'group': [(70, 'EGNQST'), (184, 'IV')], 'value': 10},
                          {'group': [(65, 'R'), (68, 'NGR')], 'value': 5},
                          {'group': [(62, 'V'), (65, 'R')], 'value': 5},
                          {'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                          {'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                          {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                          {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 10},
                          {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                           'value': 5},
                          {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')], 'value': 5},
                          {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 10}],
                         {1: {'max': 9, 'min': -inf},
                          2: {'max': 14, 'min': 10},
                          3: {'max': 29, 'min': 15},
                          4: {'max': 59, 'min': 30},
                          5: {'max': inf, 'min': 60}}),
                 'TPV/r': ([{'group': [(24, 'I')], 'value': -5},
                            {'group': [(32, 'I')], 'value': 5},
                            {'group': [(33, 'F')], 'value': 10},
                            {'group': [(43, 'T')], 'value': 10},
                            [{'group': [(46, 'I')], 'value': 5},
                             {'group': [(46, 'L')], 'value': 10},
                             {'group': [(46, 'V')], 'value': 5}],
                            [{'group': [(47, 'A')], 'value': 30},
                             {'group': [(47, 'V')], 'value': 30}],
                            [{'group': [(50, 'L')], 'value': -5},
                             {'group': [(50, 'V')], 'value': -5}],
                            [{'group': [(54, 'A')], 'value': 20},
                             {'group': [(54, 'L')], 'value': -10},
                             {'group': [(54, 'M')], 'value': 20},
                             {'group': [(54, 'S')], 'value': 20},
                             {'group': [(54, 'T')], 'value': 20},
                             {'group': [(54, 'V')], 'value': 20}],
                            {'group': [(58, 'E')], 'value': 15},
                            {'group': [(74, 'P')], 'value': 25},
                            {'group': [(76, 'V')], 'value': -5},
                            [{'group': [(82, 'C')], 'value': 10},
                             {'group': [(82, 'L')], 'value': 45},
                             {'group': [(82, 'M')], 'value': 10},
                             {'group': [(82, 'S')], 'value': 30},
                             {'group': [(82, 'T')], 'value': 45}],
                            {'group': [(83, 'D')], 'value': 25},
                            [{'group': [(84, 'A')], 'value': 60},
                             {'group': [(84, 'C')], 'value': 30},
                             {'group': [(84, 'V')], 'value': 30}],
                            {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5}],
                           {1: {'max': 9, 'min': -inf},
                            2: {'max': 14, 'min': 10},
                            3: {'max': 29, 'min': 15},
                            4: {'max': 59, 'min': 30},
                            5: {'max': inf, 'min': 60}}),
                 'abacavir': ([{'group': [(41, 'L')], 'value': 5},
                               {'group': [(62, 'V')], 'value': 5},
                               [{'group': [(65, 'E')], 'value': 10},
                                {'group': [(65, 'N')], 'value': 30},
                                {'group': [(65, 'R')], 'value': 45}],
                               [{'group': [(67, 'E')], 'value': 5},
                                {'group': [(67, 'G')], 'value': 5},
                                {'group': [(67, 'H')], 'value': 5},
                                {'group': [(67, 'N')], 'value': 5},
                                {'group': [(67, 'S')], 'value': 5},
                                {'group': [(67, 'T')], 'value': 5},
                                {'group': [(67, 'd')], 'value': 30}],
                               {'group': [(68, 'd')], 'value': 15},
                               [{'group': [(69, 'G')], 'value': 10},
                                {'group': [(69, 'i')], 'value': 60},
                                {'group': [(69, 'd')], 'value': 15}],
                               [{'group': [(70, 'E')], 'value': 15},
                                {'group': [(70, 'G')], 'value': 15},
                                {'group': [(70, 'N')], 'value': 15},
                                {'group': [(70, 'Q')], 'value': 15},
                                {'group': [(70, 'R')], 'value': 5},
                                {'group': [(70, 'S')], 'value': 15},
                                {'group': [(70, 'T')], 'value': 15},
                                {'group': [(70, 'd')], 'value': 15}],
                               [{'group': [(74, 'I')], 'value': 15},
                                {'group': [(74, 'V')], 'value': 30}],
                               {'group': [(75, 'I')], 'value': 5},
                               {'group': [(77, 'L')], 'value': 5},
                               {'group': [(115, 'F')], 'value': 30},
                               {'group': [(116, 'Y')], 'value': 5},
                               [{'group': [(151, 'L')], 'value': 30},
                                {'group': [(151, 'M')], 'value': 60}],
                               [{'group': [(184, 'I')], 'value': 15},
                                {'group': [(184, 'V')], 'value': 15}],
                               {'group': [(210, 'W')], 'value': 5},
                               [{'group': [(215, 'F')], 'value': 10},
                                {'group': [(215, 'I')], 'value': 5},
                                {'group': [(215, 'V')], 'value': 5},
                                {'group': [(215, 'Y')], 'value': 10}],
                               [{'group': [(219, 'E')], 'value': 5},
                                {'group': [(219, 'N')], 'value': 5},
                                {'group': [(219, 'Q')], 'value': 5},
                                {'group': [(219, 'R')], 'value': 5},
                                {'group': [(219, 'W')], 'value': 5}],
                               {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                               {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                               {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                               {'group': [(74, 'V'), (184, 'IV')], 'value': 15},
                               {'group': [(115, 'F'), (184, 'IV')], 'value': 15},
                               {'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                               {'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                               {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                               {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 10},
                               {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                                'value': 5},
                               {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                                'value': 5},
                               {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                               {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')],
                                'value': 5},
                               {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')],
                                'value': 10},
                               {'group': [(41, 'L'), (184, 'VI'), (215, 'FY')], 'value': 10},
                               {'group': [(67, 'EGNHST'),
                                          (70, 'R'),
                                          (184, 'VI'),
                                          (219, 'ENQRW')],
                                'value': 10}],
                              {1: {'max': 9, 'min': -inf},
                               2: {'max': 14, 'min': 10},
                               3: {'max': 29, 'min': 15},
                               4: {'max': 59, 'min': 30},
                               5: {'max': inf, 'min': 60}}),
                 'atazanavir/r': ([{'group': [(20, 'T')], 'value': 5},
                                   [{'group': [(24, 'F')], 'value': 5},
                                    {'group': [(24, 'I')], 'value': 10},
                                    {'group': [(24, 'M')], 'value': 5}],
                                   {'group': [(32, 'I')], 'value': 15},
                                   {'group': [(33, 'F')], 'value': 5},
                                   [{'group': [(46, 'I')], 'value': 10},
                                    {'group': [(46, 'L')], 'value': 10},
                                    {'group': [(46, 'V')], 'value': 10}],
                                   {'group': [(47, 'V')], 'value': 10},
                                   [{'group': [(48, 'A')], 'value': 10},
                                    {'group': [(48, 'L')], 'value': 10},
                                    {'group': [(48, 'M')], 'value': 30},
                                    {'group': [(48, 'Q')], 'value': 10},
                                    {'group': [(48, 'S')], 'value': 10},
                                    {'group': [(48, 'T')], 'value': 10},
                                    {'group': [(48, 'V')], 'value': 30}],
                                   {'group': [(50, 'L')], 'value': 60},
                                   {'group': [(53, 'L')], 'value': 10},
                                   [{'group': [(54, 'A')], 'value': 15},
                                    {'group': [(54, 'L')], 'value': 15},
                                    {'group': [(54, 'M')], 'value': 15},
                                    {'group': [(54, 'S')], 'value': 15},
                                    {'group': [(54, 'T')], 'value': 15},
                                    {'group': [(54, 'V')], 'value': 15}],
                                   [{'group': [(73, 'A')], 'value': 10},
                                    {'group': [(73, 'C')], 'value': 10},
                                    {'group': [(73, 'D')], 'value': 5},
                                    {'group': [(73, 'S')], 'value': 10},
                                    {'group': [(73, 'T')], 'value': 10},
                                    {'group': [(73, 'V')], 'value': 5}],
                                   {'group': [(74, 'P')], 'value': 10},
                                   [{'group': [(82, 'A')], 'value': 15},
                                    {'group': [(82, 'C')], 'value': 15},
                                    {'group': [(82, 'F')], 'value': 15},
                                    {'group': [(82, 'L')], 'value': 10},
                                    {'group': [(82, 'M')], 'value': 10},
                                    {'group': [(82, 'S')], 'value': 30},
                                    {'group': [(82, 'T')], 'value': 30}],
                                   {'group': [(83, 'D')], 'value': 10},
                                   [{'group': [(84, 'A')], 'value': 60},
                                    {'group': [(84, 'C')], 'value': 60},
                                    {'group': [(84, 'V')], 'value': 60}],
                                   [{'group': [(88, 'D')], 'value': 10},
                                    {'group': [(88, 'G')], 'value': 15},
                                    {'group': [(88, 'S')], 'value': 60},
                                    {'group': [(88, 'T')], 'value': 15}],
                                   {'group': [(90, 'M')], 'value': 25},
                                   {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                   {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                   {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                   {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                   {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                                   {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                   {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                                   {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                   {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                                   {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                                   {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'bictegravir': ([{'group': [(51, 'Y')], 'value': 10},
                                  [{'group': [(66, 'I')], 'value': 5},
                                   {'group': [(66, 'K')], 'value': 15}],
                                  {'group': [(92, 'Q')], 'value': 10},
                                  {'group': [(118, 'R')], 'value': 30},
                                  [{'group': [(121, 'C')], 'value': 15},
                                   {'group': [(121, 'Y')], 'value': 10}],
                                  [{'group': [(138, 'A')], 'value': 10},
                                   {'group': [(138, 'K')], 'value': 10},
                                   {'group': [(138, 'T')], 'value': 10}],
                                  [{'group': [(140, 'A')], 'value': 10},
                                   {'group': [(140, 'C')], 'value': 10},
                                   {'group': [(140, 'R')], 'value': 10},
                                   {'group': [(140, 'S')], 'value': 10}],
                                  [{'group': [(143, 'A')], 'value': 5},
                                   {'group': [(143, 'C')], 'value': 5},
                                   {'group': [(143, 'G')], 'value': 5},
                                   {'group': [(143, 'H')], 'value': 5},
                                   {'group': [(143, 'K')], 'value': 5},
                                   {'group': [(143, 'R')], 'value': 5},
                                   {'group': [(143, 'S')], 'value': 5}],
                                  {'group': [(147, 'G')], 'value': 10},
                                  [{'group': [(148, 'H')], 'value': 25},
                                   {'group': [(148, 'K')], 'value': 30},
                                   {'group': [(148, 'R')], 'value': 25}],
                                  {'group': [(151, 'L')], 'value': 15},
                                  [{'group': [(153, 'F')], 'value': 15},
                                   {'group': [(153, 'Y')], 'value': 15}],
                                  {'group': [(155, 'H')], 'value': 10},
                                  {'group': [(230, 'R')], 'value': 10},
                                  {'group': [(263, 'K')], 'value': 30},
                                  {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                                  {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 10},
                                  {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 10},
                                  {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 10},
                                  {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')],
                                   'value': 10},
                                  {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                                  {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                                  {'group': [(147, 'G'), (148, 'HKR')], 'value': 15},
                                  {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                                  {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                                  {'group': [(148, 'HKR'), (163, 'KR')], 'value': 5},
                                  {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                                  {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                                  {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                                  {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                                  {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                                  {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                                  {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                                  {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                                  {'group': [(97, 'A'), (148, 'HKR')], 'value': 15}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'cabotegravir': ([{'group': [(51, 'Y')], 'value': 15},
                                   [{'group': [(66, 'I')], 'value': 10},
                                    {'group': [(66, 'K')], 'value': 20}],
                                   [{'group': [(92, 'G')], 'value': 10},
                                    {'group': [(92, 'Q')], 'value': 15},
                                    {'group': [(92, 'V')], 'value': 10}],
                                   {'group': [(118, 'R')], 'value': 60},
                                   [{'group': [(121, 'C')], 'value': 60},
                                    {'group': [(121, 'Y')], 'value': 15}],
                                   [{'group': [(138, 'A')], 'value': 10},
                                    {'group': [(138, 'K')], 'value': 10},
                                    {'group': [(138, 'T')], 'value': 10}],
                                   [{'group': [(140, 'A')], 'value': 10},
                                    {'group': [(140, 'C')], 'value': 10},
                                    {'group': [(140, 'R')], 'value': 60},
                                    {'group': [(140, 'S')], 'value': 10}],
                                   [{'group': [(143, 'A')], 'value': 10},
                                    {'group': [(143, 'C')], 'value': 10},
                                    {'group': [(143, 'G')], 'value': 10},
                                    {'group': [(143, 'H')], 'value': 10},
                                    {'group': [(143, 'K')], 'value': 10},
                                    {'group': [(143, 'R')], 'value': 10},
                                    {'group': [(143, 'S')], 'value': 10}],
                                   {'group': [(146, 'P')], 'value': 30},
                                   {'group': [(147, 'G')], 'value': 10},
                                   [{'group': [(148, 'H')], 'value': 30},
                                    {'group': [(148, 'K')], 'value': 50},
                                    {'group': [(148, 'N')], 'value': 15},
                                    {'group': [(148, 'R')], 'value': 40}],
                                   {'group': [(151, 'L')], 'value': 15},
                                   [{'group': [(153, 'F')], 'value': 25},
                                    {'group': [(153, 'Y')], 'value': 25}],
                                   [{'group': [(155, 'H')], 'value': 25},
                                    {'group': [(155, 'S')], 'value': 15},
                                    {'group': [(155, 'T')], 'value': 15}],
                                   {'group': [(230, 'R')], 'value': 20},
                                   {'group': [(263, 'K')], 'value': 30},
                                   {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                                   {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                                   {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 20},
                                   {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 20},
                                   {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')],
                                    'value': 10},
                                   {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                                   {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                                   {'group': [(147, 'G'), (148, 'HKR')], 'value': 20},
                                   {'group': [(147, 'G'), (155, 'H')], 'value': 25},
                                   {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                                   {'group': [(148, 'HKR'), (163, 'KR')], 'value': 20},
                                   {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                                   {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                                   {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                                   {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                                   {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 10},
                                   {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                                   {'group': [(92, 'Q'), (155, 'H')], 'value': 20},
                                   {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                                   {'group': [(97, 'A'), (143, 'ACGHRS')], 'value': 5},
                                   {'group': [(97, 'A'), (148, 'HKR')], 'value': 20}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'darunavir/r': ([{'group': [(10, 'F')], 'value': 5},
                                  {'group': [(32, 'I')], 'value': 15},
                                  {'group': [(33, 'F')], 'value': 5},
                                  [{'group': [(47, 'A')], 'value': 10},
                                   {'group': [(47, 'V')], 'value': 10}],
                                  [{'group': [(50, 'L')], 'value': -10},
                                   {'group': [(50, 'V')], 'value': 20}],
                                  [{'group': [(54, 'L')], 'value': 20},
                                   {'group': [(54, 'M')], 'value': 20}],
                                  {'group': [(74, 'P')], 'value': 5},
                                  {'group': [(76, 'V')], 'value': 20},
                                  {'group': [(82, 'F')], 'value': 15},
                                  [{'group': [(84, 'A')], 'value': 30},
                                   {'group': [(84, 'C')], 'value': 15},
                                   {'group': [(84, 'V')], 'value': 15}],
                                  {'group': [(88, 'S')], 'value': -5},
                                  {'group': [(89, 'V')], 'value': 5},
                                  {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                                  {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                  {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                                  {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                  {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                                  {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                                  {'group': [(54, 'LM'), (89, 'V')], 'value': 5}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'didanosine': ([{'group': [(41, 'L')], 'value': 10},
                                 {'group': [(62, 'V')], 'value': 5},
                                 [{'group': [(65, 'E')], 'value': 10},
                                  {'group': [(65, 'N')], 'value': 30},
                                  {'group': [(65, 'R')], 'value': 60}],
                                 [{'group': [(67, 'E')], 'value': 5},
                                  {'group': [(67, 'G')], 'value': 5},
                                  {'group': [(67, 'H')], 'value': 5},
                                  {'group': [(67, 'N')], 'value': 5},
                                  {'group': [(67, 'S')], 'value': 5},
                                  {'group': [(67, 'T')], 'value': 5},
                                  {'group': [(67, 'd')], 'value': 30}],
                                 {'group': [(68, 'd')], 'value': 30},
                                 [{'group': [(69, 'G')], 'value': 10},
                                  {'group': [(69, 'i')], 'value': 60},
                                  {'group': [(69, 'd')], 'value': 30},
                                  {'group': [(69, 'D')], 'value': 30}],
                                 [{'group': [(70, 'E')], 'value': 15},
                                  {'group': [(70, 'G')], 'value': 15},
                                  {'group': [(70, 'N')], 'value': 15},
                                  {'group': [(70, 'Q')], 'value': 15},
                                  {'group': [(70, 'R')], 'value': 10},
                                  {'group': [(70, 'S')], 'value': 15},
                                  {'group': [(70, 'T')], 'value': 15},
                                  {'group': [(70, 'd')], 'value': 30}],
                                 [{'group': [(74, 'I')], 'value': 60},
                                  {'group': [(74, 'V')], 'value': 60}],
                                 [{'group': [(75, 'I')], 'value': 5},
                                  {'group': [(75, 'M')], 'value': 15},
                                  {'group': [(75, 'A')], 'value': 15},
                                  {'group': [(75, 'S')], 'value': 15},
                                  {'group': [(75, 'T')], 'value': 30}],
                                 {'group': [(77, 'L')], 'value': 10},
                                 {'group': [(116, 'Y')], 'value': 10},
                                 [{'group': [(151, 'L')], 'value': 30},
                                  {'group': [(151, 'M')], 'value': 60}],
                                 [{'group': [(184, 'I')], 'value': 10},
                                  {'group': [(184, 'V')], 'value': 10}],
                                 {'group': [(210, 'W')], 'value': 10},
                                 [{'group': [(215, 'A')], 'value': 10},
                                  {'group': [(215, 'C')], 'value': 10},
                                  {'group': [(215, 'D')], 'value': 10},
                                  {'group': [(215, 'E')], 'value': 10},
                                  {'group': [(215, 'F')], 'value': 15},
                                  {'group': [(215, 'I')], 'value': 10},
                                  {'group': [(215, 'L')], 'value': 10},
                                  {'group': [(215, 'N')], 'value': 10},
                                  {'group': [(215, 'S')], 'value': 10},
                                  {'group': [(215, 'V')], 'value': 10},
                                  {'group': [(215, 'Y')], 'value': 15}],
                                 [{'group': [(219, 'E')], 'value': 5},
                                  {'group': [(219, 'N')], 'value': 5},
                                  {'group': [(219, 'Q')], 'value': 5},
                                  {'group': [(219, 'R')], 'value': 5},
                                  {'group': [(219, 'W')], 'value': 5}],
                                 {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                                  'value': 5},
                                 {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                                 {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                                  'value': 5},
                                 {'group': [(41, 'L'), (67, 'EGN'), (215, 'FY')], 'value': 5},
                                 {'group': [(67, 'EGN'), (215, 'FY'), (219, 'ENQR')],
                                  'value': 5},
                                 {'group': [(67, 'EGN'), (70, 'R'), (219, 'ENQR')],
                                  'value': 10},
                                 {'group': [(70, 'R'), (215, 'FY')], 'value': 5},
                                 {'group': [(77, 'L'), (116, 'Y'), (151, 'M')], 'value': 10},
                                 [{'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 5},
                                  {'group': [(210, 'W'), (215, 'FY')], 'value': 10}],
                                 [{'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 5},
                                  {'group': [(41, 'L'), (215, 'FY')], 'value': 10}]],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'dolutegravir': ([{'group': [(51, 'Y')], 'value': 10},
                                   [{'group': [(66, 'I')], 'value': 5},
                                    {'group': [(66, 'K')], 'value': 15}],
                                   {'group': [(92, 'Q')], 'value': 10},
                                   {'group': [(118, 'R')], 'value': 50},
                                   [{'group': [(121, 'C')], 'value': 15},
                                    {'group': [(121, 'Y')], 'value': 10}],
                                   [{'group': [(138, 'A')], 'value': 10},
                                    {'group': [(138, 'K')], 'value': 10},
                                    {'group': [(138, 'T')], 'value': 10}],
                                   [{'group': [(140, 'A')], 'value': 10},
                                    {'group': [(140, 'C')], 'value': 10},
                                    {'group': [(140, 'R')], 'value': 10},
                                    {'group': [(140, 'S')], 'value': 10}],
                                   [{'group': [(143, 'A')], 'value': 5},
                                    {'group': [(143, 'C')], 'value': 5},
                                    {'group': [(143, 'G')], 'value': 5},
                                    {'group': [(143, 'H')], 'value': 5},
                                    {'group': [(143, 'K')], 'value': 5},
                                    {'group': [(143, 'R')], 'value': 5},
                                    {'group': [(143, 'S')], 'value': 5}],
                                   {'group': [(147, 'G')], 'value': 10},
                                   [{'group': [(148, 'H')], 'value': 25},
                                    {'group': [(148, 'K')], 'value': 30},
                                    {'group': [(148, 'R')], 'value': 25}],
                                   {'group': [(151, 'L')], 'value': 15},
                                   [{'group': [(153, 'F')], 'value': 15},
                                    {'group': [(153, 'Y')], 'value': 15}],
                                   {'group': [(155, 'H')], 'value': 10},
                                   {'group': [(230, 'R')], 'value': 20},
                                   {'group': [(263, 'K')], 'value': 30},
                                   {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                                   {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 10},
                                   {'group': [(138, 'AKT'), (148, 'HKR')], 'value': 10},
                                   {'group': [(140, 'ACS'), (148, 'HKR')], 'value': 10},
                                   {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')],
                                    'value': 10},
                                   {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                                   {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                                   {'group': [(147, 'G'), (148, 'HKR')], 'value': 15},
                                   {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                                   {'group': [(148, 'HKR'), (155, 'H')], 'value': 20},
                                   {'group': [(148, 'HKR'), (163, 'KR')], 'value': 5},
                                   {'group': [(155, 'H'), (263, 'K')], 'value': 20},
                                   {'group': [(157, 'Q'), (263, 'K')], 'value': 10},
                                   {'group': [(51, 'Y'), (263, 'K')], 'value': 10},
                                   {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                                   {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                                   {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                                   {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                                   {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                                   {'group': [(97, 'A'), (148, 'HKR')], 'value': 15}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'doravirine': ([{'group': [(98, 'G')], 'value': 15},
                                 [{'group': [(100, 'I')], 'value': 15},
                                  {'group': [(100, 'V')], 'value': 10}],
                                 [{'group': [(101, 'E')], 'value': 15},
                                  {'group': [(101, 'P')], 'value': 10}],
                                 [{'group': [(106, 'A')], 'value': 60},
                                  {'group': [(106, 'I')], 'value': 10},
                                  {'group': [(106, 'M')], 'value': 30}],
                                 {'group': [(108, 'I')], 'value': 10},
                                 {'group': [(138, 'K')], 'value': 5},
                                 {'group': [(179, 'F')], 'value': 10},
                                 [{'group': [(181, 'C')], 'value': 10},
                                  {'group': [(181, 'I')], 'value': 20},
                                  {'group': [(181, 'V')], 'value': 20}],
                                 [{'group': [(188, 'C')], 'value': 5},
                                  {'group': [(188, 'F')], 'value': 30},
                                  {'group': [(188, 'H')], 'value': 5},
                                  {'group': [(188, 'L')], 'value': 60}],
                                 [{'group': [(190, 'C')], 'value': 10},
                                  {'group': [(190, 'E')], 'value': 60},
                                  {'group': [(190, 'Q')], 'value': 30},
                                  {'group': [(190, 'S')], 'value': 20},
                                  {'group': [(190, 'T')], 'value': 10},
                                  {'group': [(190, 'V')], 'value': 10}],
                                 {'group': [(221, 'Y')], 'value': 10},
                                 {'group': [(225, 'H')], 'value': 20},
                                 [{'group': [(227, 'C')], 'value': 60},
                                  {'group': [(227, 'I')], 'value': 60},
                                  {'group': [(227, 'L')], 'value': 60},
                                  {'group': [(227, 'V')], 'value': 60}],
                                 [{'group': [(230, 'I')], 'value': 15},
                                  {'group': [(230, 'L')], 'value': 60}],
                                 {'group': [(234, 'I')], 'value': 45},
                                 {'group': [(236, 'L')], 'value': 10},
                                 {'group': [(318, 'F')], 'value': 60},
                                 {'group': [(100, 'I'), (103, 'N')], 'value': 15},
                                 {'group': [(100, 'I'), (103, 'N'), (225, 'H')], 'value': -10},
                                 {'group': [(103, 'N'), (181, 'C')], 'value': 5},
                                 {'group': [(103, 'N'), (225, 'H')], 'value': 10},
                                 {'group': [(106, 'I'), (181, 'C')], 'value': 5},
                                 {'group': [(106, 'I'), (190, 'S')], 'value': 5},
                                 {'group': [(108, 'I'), (181, 'C')], 'value': 5},
                                 {'group': [(108, 'I'), (234, 'I')], 'value': 10},
                                 {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                                 {'group': [(181, 'CIV'), (221, 'Y')], 'value': 10},
                                 {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                                 {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                                 [{'group': [(101, 'E'), (190, 'A')], 'value': 5},
                                  {'group': [(101, 'E'), (190, 'S')], 'value': 5}]],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'efavirenz': ([{'group': [(98, 'G')], 'value': 15},
                                [{'group': [(100, 'I')], 'value': 60},
                                 {'group': [(100, 'V')], 'value': 30}],
                                [{'group': [(101, 'E')], 'value': 15},
                                 {'group': [(101, 'H')], 'value': 10},
                                 {'group': [(101, 'P')], 'value': 60}],
                                [{'group': [(103, 'H')], 'value': 60},
                                 {'group': [(103, 'N')], 'value': 60},
                                 {'group': [(103, 'S')], 'value': 45},
                                 {'group': [(103, 'T')], 'value': 15}],
                                [{'group': [(106, 'A')], 'value': 45},
                                 {'group': [(106, 'M')], 'value': 60}],
                                {'group': [(108, 'I')], 'value': 10},
                                [{'group': [(138, 'G')], 'value': 10},
                                 {'group': [(138, 'K')], 'value': 10},
                                 {'group': [(138, 'Q')], 'value': 10},
                                 {'group': [(138, 'R')], 'value': 10}],
                                [{'group': [(179, 'D')], 'value': 10},
                                 {'group': [(179, 'E')], 'value': 10},
                                 {'group': [(179, 'F')], 'value': 10},
                                 {'group': [(179, 'L')], 'value': 10}],
                                [{'group': [(181, 'C')], 'value': 30},
                                 {'group': [(181, 'F')], 'value': 15},
                                 {'group': [(181, 'G')], 'value': 15},
                                 {'group': [(181, 'I')], 'value': 30},
                                 {'group': [(181, 'S')], 'value': 15},
                                 {'group': [(181, 'V')], 'value': 30}],
                                [{'group': [(188, 'C')], 'value': 60},
                                 {'group': [(188, 'F')], 'value': 60},
                                 {'group': [(188, 'H')], 'value': 30},
                                 {'group': [(188, 'L')], 'value': 60}],
                                [{'group': [(190, 'A')], 'value': 45},
                                 {'group': [(190, 'C')], 'value': 60},
                                 {'group': [(190, 'E')], 'value': 60},
                                 {'group': [(190, 'Q')], 'value': 60},
                                 {'group': [(190, 'S')], 'value': 60},
                                 {'group': [(190, 'T')], 'value': 60},
                                 {'group': [(190, 'V')], 'value': 60}],
                                {'group': [(221, 'Y')], 'value': 10},
                                {'group': [(225, 'H')], 'value': 45},
                                [{'group': [(227, 'C')], 'value': 45},
                                 {'group': [(227, 'I')], 'value': 10},
                                 {'group': [(227, 'L')], 'value': 15},
                                 {'group': [(227, 'V')], 'value': 10}],
                                [{'group': [(230, 'I')], 'value': 15},
                                 {'group': [(230, 'L')], 'value': 45}],
                                [{'group': [(238, 'N')], 'value': 10},
                                 {'group': [(238, 'T')], 'value': 30}],
                                {'group': [(318, 'F')], 'value': 10},
                                {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                                {'group': [(103, 'R'), (179, 'D')], 'value': 20},
                                {'group': [(106, 'A'), (227, 'CL')], 'value': 15},
                                {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                                {'group': [(98, 'G'), (227, 'CL')], 'value': 15}],
                               {1: {'max': 9, 'min': -inf},
                                2: {'max': 14, 'min': 10},
                                3: {'max': 29, 'min': 15},
                                4: {'max': 59, 'min': 30},
                                5: {'max': inf, 'min': 60}}),
                 'elvitegravir': ([{'group': [(51, 'Y')], 'value': 15},
                                   [{'group': [(66, 'A')], 'value': 60},
                                    {'group': [(66, 'I')], 'value': 60},
                                    {'group': [(66, 'K')], 'value': 60}],
                                   [{'group': [(92, 'G')], 'value': 30},
                                    {'group': [(92, 'Q')], 'value': 60},
                                    {'group': [(92, 'V')], 'value': 60}],
                                   {'group': [(95, 'K')], 'value': 10},
                                   {'group': [(97, 'A')], 'value': 10},
                                   {'group': [(118, 'R')], 'value': 60},
                                   [{'group': [(121, 'C')], 'value': 60},
                                    {'group': [(121, 'Y')], 'value': 60}],
                                   [{'group': [(138, 'A')], 'value': 15},
                                    {'group': [(138, 'K')], 'value': 15},
                                    {'group': [(138, 'T')], 'value': 15}],
                                   [{'group': [(140, 'A')], 'value': 30},
                                    {'group': [(140, 'C')], 'value': 30},
                                    {'group': [(140, 'R')], 'value': 30},
                                    {'group': [(140, 'S')], 'value': 30}],
                                   [{'group': [(143, 'A')], 'value': 10},
                                    {'group': [(143, 'C')], 'value': 10},
                                    {'group': [(143, 'G')], 'value': 10},
                                    {'group': [(143, 'H')], 'value': 10},
                                    {'group': [(143, 'K')], 'value': 10},
                                    {'group': [(143, 'R')], 'value': 10},
                                    {'group': [(143, 'S')], 'value': 10}],
                                   {'group': [(145, 'S')], 'value': 60},
                                   {'group': [(146, 'P')], 'value': 60},
                                   {'group': [(147, 'G')], 'value': 60},
                                   [{'group': [(148, 'H')], 'value': 60},
                                    {'group': [(148, 'K')], 'value': 60},
                                    {'group': [(148, 'N')], 'value': 15},
                                    {'group': [(148, 'R')], 'value': 60}],
                                   [{'group': [(151, 'A')], 'value': 30},
                                    {'group': [(151, 'L')], 'value': 60}],
                                   [{'group': [(153, 'F')], 'value': 25},
                                    {'group': [(153, 'Y')], 'value': 25}],
                                   [{'group': [(155, 'H')], 'value': 60},
                                    {'group': [(155, 'S')], 'value': 30},
                                    {'group': [(155, 'T')], 'value': 30}],
                                   {'group': [(157, 'Q')], 'value': 10},
                                   [{'group': [(163, 'K')], 'value': 15},
                                    {'group': [(163, 'R')], 'value': 15}],
                                   {'group': [(230, 'R')], 'value': 20},
                                   {'group': [(232, 'N')], 'value': 10},
                                   {'group': [(263, 'K')], 'value': 30},
                                   {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                                   {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                                   {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')],
                                    'value': 10},
                                   {'group': [(143, 'ACGHRS'), (163, 'R')], 'value': 5},
                                   {'group': [(143, 'ACGHRS'), (230, 'R')], 'value': 5},
                                   {'group': [(147, 'G'), (155, 'H')], 'value': 10},
                                   {'group': [(51, 'Y'), (263, 'K')], 'value': 15},
                                   {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                                   {'group': [(74, 'FIM'), (143, 'ACGHRS')], 'value': 5},
                                   {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                                   {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                                   {'group': [(97, 'A'), (118, 'R')], 'value': 10},
                                   {'group': [(97, 'A'), (143, 'ACGHRS')], 'value': 5}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'emtricitabine': ([[{'group': [(65, 'N')], 'value': 15},
                                     {'group': [(65, 'R')], 'value': 30}],
                                    {'group': [(67, 'd')], 'value': 15},
                                    {'group': [(68, 'd')], 'value': 15},
                                    [{'group': [(69, 'i')], 'value': 30},
                                     {'group': [(69, 'd')], 'value': 15}],
                                    [{'group': [(70, 'E')], 'value': 10},
                                     {'group': [(70, 'G')], 'value': 10},
                                     {'group': [(70, 'N')], 'value': 10},
                                     {'group': [(70, 'Q')], 'value': 10},
                                     {'group': [(70, 'S')], 'value': 10},
                                     {'group': [(70, 'T')], 'value': 10},
                                     {'group': [(70, 'd')], 'value': 10}],
                                    {'group': [(75, 'I')], 'value': 5},
                                    {'group': [(77, 'L')], 'value': 5},
                                    {'group': [(116, 'Y')], 'value': 5},
                                    [{'group': [(151, 'L')], 'value': 10},
                                     {'group': [(151, 'M')], 'value': 15}],
                                    [{'group': [(184, 'I')], 'value': 60},
                                     {'group': [(184, 'V')], 'value': 60}],
                                    {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                                    {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')],
                                     'value': 15},
                                    {'group': [(41, 'L'), (215, 'FY')], 'value': 5},
                                    {'group': [(41, 'L'), (210, 'W'), (215, 'FY')],
                                     'value': 15},
                                    {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')],
                                     'value': 10}],
                                   {1: {'max': 9, 'min': -inf},
                                    2: {'max': 14, 'min': 10},
                                    3: {'max': 29, 'min': 15},
                                    4: {'max': 59, 'min': 30},
                                    5: {'max': inf, 'min': 60}}),
                 'etravirine': ([{'group': [(98, 'G')], 'value': 10},
                                 [{'group': [(100, 'I')], 'value': 30},
                                  {'group': [(100, 'V')], 'value': 10}],
                                 [{'group': [(101, 'E')], 'value': 15},
                                  {'group': [(101, 'H')], 'value': 10},
                                  {'group': [(101, 'P')], 'value': 60}],
                                 {'group': [(106, 'I')], 'value': 10},
                                 [{'group': [(138, 'A')], 'value': 10},
                                  {'group': [(138, 'G')], 'value': 10},
                                  {'group': [(138, 'K')], 'value': 10},
                                  {'group': [(138, 'Q')], 'value': 10},
                                  {'group': [(138, 'R')], 'value': 10}],
                                 [{'group': [(179, 'D')], 'value': 10},
                                  {'group': [(179, 'E')], 'value': 10},
                                  {'group': [(179, 'F')], 'value': 15},
                                  {'group': [(179, 'L')], 'value': 10}],
                                 [{'group': [(181, 'C')], 'value': 30},
                                  {'group': [(181, 'F')], 'value': 15},
                                  {'group': [(181, 'G')], 'value': 15},
                                  {'group': [(181, 'I')], 'value': 60},
                                  {'group': [(181, 'S')], 'value': 15},
                                  {'group': [(181, 'V')], 'value': 60}],
                                 {'group': [(188, 'L')], 'value': 10},
                                 [{'group': [(190, 'A')], 'value': 10},
                                  {'group': [(190, 'C')], 'value': 10},
                                  {'group': [(190, 'E')], 'value': 45},
                                  {'group': [(190, 'Q')], 'value': 45},
                                  {'group': [(190, 'S')], 'value': 10},
                                  {'group': [(190, 'T')], 'value': 10},
                                  {'group': [(190, 'V')], 'value': 10}],
                                 {'group': [(221, 'Y')], 'value': 10},
                                 {'group': [(227, 'C')], 'value': 30},
                                 [{'group': [(230, 'I')], 'value': 15},
                                  {'group': [(230, 'L')], 'value': 30}],
                                 {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                                 {'group': [(101, 'E'), (188, 'L')], 'value': 5},
                                 {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                                 {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                                 {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                                 [{'group': [(101, 'E'), (190, 'A')], 'value': 5},
                                  {'group': [(101, 'E'), (190, 'S')], 'value': 5}],
                                 [{'group': [(179, 'F'), (181, 'C')], 'value': 15},
                                  {'group': [(179, 'T'), (181, 'C')], 'value': 10}]],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'fosamprenavir/r': ([{'group': [(10, 'F')], 'value': 15},
                                      {'group': [(20, 'T')], 'value': 5},
                                      [{'group': [(24, 'F')], 'value': 5},
                                       {'group': [(24, 'I')], 'value': 10},
                                       {'group': [(24, 'M')], 'value': 5}],
                                      {'group': [(32, 'I')], 'value': 30},
                                      {'group': [(33, 'F')], 'value': 10},
                                      [{'group': [(46, 'I')], 'value': 10},
                                       {'group': [(46, 'L')], 'value': 10},
                                       {'group': [(46, 'V')], 'value': 10}],
                                      [{'group': [(47, 'A')], 'value': 60},
                                       {'group': [(47, 'V')], 'value': 35}],
                                      [{'group': [(50, 'L')], 'value': -5},
                                       {'group': [(50, 'V')], 'value': 60}],
                                      [{'group': [(54, 'A')], 'value': 10},
                                       {'group': [(54, 'L')], 'value': 60},
                                       {'group': [(54, 'M')], 'value': 60},
                                       {'group': [(54, 'S')], 'value': 10},
                                       {'group': [(54, 'T')], 'value': 10},
                                       {'group': [(54, 'V')], 'value': 10}],
                                      [{'group': [(73, 'A')], 'value': 10},
                                       {'group': [(73, 'C')], 'value': 10},
                                       {'group': [(73, 'D')], 'value': 5},
                                       {'group': [(73, 'S')], 'value': 10},
                                       {'group': [(73, 'T')], 'value': 10},
                                       {'group': [(73, 'V')], 'value': 5}],
                                      {'group': [(74, 'P')], 'value': 10},
                                      {'group': [(76, 'V')], 'value': 60},
                                      [{'group': [(82, 'A')], 'value': 15},
                                       {'group': [(82, 'C')], 'value': 15},
                                       {'group': [(82, 'F')], 'value': 30},
                                       {'group': [(82, 'L')], 'value': 15},
                                       {'group': [(82, 'M')], 'value': 15},
                                       {'group': [(82, 'S')], 'value': 15},
                                       {'group': [(82, 'T')], 'value': 15}],
                                      [{'group': [(84, 'A')], 'value': 60},
                                       {'group': [(84, 'C')], 'value': 60},
                                       {'group': [(84, 'V')], 'value': 60}],
                                      {'group': [(88, 'S')], 'value': -10},
                                      {'group': [(89, 'V')], 'value': 10},
                                      {'group': [(90, 'M')], 'value': 20},
                                      {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                                      {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                                      {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                      {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                      {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                                      {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                                      {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                                      {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                      {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                                      {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                      {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                                      {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                      {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                                      {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                                      {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                      {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                                      {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                                      {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                                      {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                                      {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                                     {1: {'max': 9, 'min': -inf},
                                      2: {'max': 14, 'min': 10},
                                      3: {'max': 29, 'min': 15},
                                      4: {'max': 59, 'min': 30},
                                      5: {'max': inf, 'min': 60}}),
                 'indinavir/r': ([{'group': [(10, 'F')], 'value': 10},
                                  {'group': [(20, 'T')], 'value': 5},
                                  [{'group': [(24, 'F')], 'value': 5},
                                   {'group': [(24, 'I')], 'value': 15},
                                   {'group': [(24, 'M')], 'value': 5}],
                                  {'group': [(32, 'I')], 'value': 15},
                                  {'group': [(33, 'F')], 'value': 5},
                                  [{'group': [(46, 'I')], 'value': 10},
                                   {'group': [(46, 'L')], 'value': 10},
                                   {'group': [(46, 'V')], 'value': 10}],
                                  [{'group': [(47, 'A')], 'value': 15},
                                   {'group': [(47, 'V')], 'value': 15}],
                                  [{'group': [(48, 'A')], 'value': 10},
                                   {'group': [(48, 'L')], 'value': 10},
                                   {'group': [(48, 'M')], 'value': 10},
                                   {'group': [(48, 'Q')], 'value': 10},
                                   {'group': [(48, 'S')], 'value': 10},
                                   {'group': [(48, 'T')], 'value': 10},
                                   {'group': [(48, 'V')], 'value': 10}],
                                  {'group': [(50, 'L')], 'value': -5},
                                  [{'group': [(54, 'A')], 'value': 15},
                                   {'group': [(54, 'L')], 'value': 10},
                                   {'group': [(54, 'M')], 'value': 15},
                                   {'group': [(54, 'S')], 'value': 15},
                                   {'group': [(54, 'T')], 'value': 15},
                                   {'group': [(54, 'V')], 'value': 15}],
                                  [{'group': [(73, 'A')], 'value': 15},
                                   {'group': [(73, 'C')], 'value': 15},
                                   {'group': [(73, 'D')], 'value': 5},
                                   {'group': [(73, 'S')], 'value': 15},
                                   {'group': [(73, 'T')], 'value': 15},
                                   {'group': [(73, 'V')], 'value': 5}],
                                  {'group': [(74, 'P')], 'value': 10},
                                  {'group': [(76, 'V')], 'value': 30},
                                  [{'group': [(82, 'A')], 'value': 30},
                                   {'group': [(82, 'C')], 'value': 15},
                                   {'group': [(82, 'F')], 'value': 30},
                                   {'group': [(82, 'L')], 'value': 10},
                                   {'group': [(82, 'M')], 'value': 30},
                                   {'group': [(82, 'S')], 'value': 30},
                                   {'group': [(82, 'T')], 'value': 30}],
                                  {'group': [(83, 'D')], 'value': 10},
                                  [{'group': [(84, 'A')], 'value': 60},
                                   {'group': [(84, 'C')], 'value': 60},
                                   {'group': [(84, 'V')], 'value': 60}],
                                  {'group': [(88, 'S')], 'value': 15},
                                  {'group': [(89, 'V')], 'value': 5},
                                  {'group': [(90, 'M')], 'value': 30},
                                  {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                                  {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                  {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                                  {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                  {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                                  {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                  {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                                  {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                  {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                                  {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                                  {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                  {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                                  {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                                  {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                                  {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                                  {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'lamivudine': ([[{'group': [(65, 'N')], 'value': 15},
                                  {'group': [(65, 'R')], 'value': 30}],
                                 {'group': [(67, 'd')], 'value': 15},
                                 {'group': [(68, 'd')], 'value': 15},
                                 [{'group': [(69, 'i')], 'value': 30},
                                  {'group': [(69, 'd')], 'value': 15}],
                                 [{'group': [(70, 'E')], 'value': 10},
                                  {'group': [(70, 'G')], 'value': 10},
                                  {'group': [(70, 'N')], 'value': 10},
                                  {'group': [(70, 'Q')], 'value': 10},
                                  {'group': [(70, 'S')], 'value': 10},
                                  {'group': [(70, 'T')], 'value': 10},
                                  {'group': [(70, 'd')], 'value': 10}],
                                 {'group': [(75, 'I')], 'value': 5},
                                 {'group': [(77, 'L')], 'value': 5},
                                 {'group': [(116, 'Y')], 'value': 5},
                                 [{'group': [(151, 'L')], 'value': 10},
                                  {'group': [(151, 'M')], 'value': 15}],
                                 [{'group': [(184, 'I')], 'value': 60},
                                  {'group': [(184, 'V')], 'value': 60}],
                                 {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                                 {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                                 {'group': [(41, 'L'), (215, 'FY')], 'value': 5},
                                 {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 15},
                                 {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')],
                                  'value': 10}],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'lenacapavir': ([{'group': [(9999, 'A')], 'value': 0}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'lopinavir/r': ([{'group': [(10, 'F')], 'value': 5},
                                  [{'group': [(24, 'F')], 'value': 5},
                                   {'group': [(24, 'I')], 'value': 10},
                                   {'group': [(24, 'M')], 'value': 5}],
                                  {'group': [(32, 'I')], 'value': 15},
                                  {'group': [(33, 'F')], 'value': 5},
                                  [{'group': [(46, 'I')], 'value': 10},
                                   {'group': [(46, 'L')], 'value': 10},
                                   {'group': [(46, 'V')], 'value': 5}],
                                  [{'group': [(47, 'A')], 'value': 60},
                                   {'group': [(47, 'V')], 'value': 15}],
                                  [{'group': [(48, 'A')], 'value': 10},
                                   {'group': [(48, 'L')], 'value': 10},
                                   {'group': [(48, 'M')], 'value': 10},
                                   {'group': [(48, 'Q')], 'value': 10},
                                   {'group': [(48, 'S')], 'value': 10},
                                   {'group': [(48, 'T')], 'value': 10},
                                   {'group': [(48, 'V')], 'value': 10}],
                                  [{'group': [(50, 'L')], 'value': -10},
                                   {'group': [(50, 'V')], 'value': 30}],
                                  [{'group': [(54, 'A')], 'value': 15},
                                   {'group': [(54, 'L')], 'value': 20},
                                   {'group': [(54, 'M')], 'value': 20},
                                   {'group': [(54, 'S')], 'value': 15},
                                   {'group': [(54, 'T')], 'value': 15},
                                   {'group': [(54, 'V')], 'value': 15}],
                                  [{'group': [(73, 'A')], 'value': 5},
                                   {'group': [(73, 'C')], 'value': 5},
                                   {'group': [(73, 'D')], 'value': 5},
                                   {'group': [(73, 'S')], 'value': 5},
                                   {'group': [(73, 'T')], 'value': 5},
                                   {'group': [(73, 'V')], 'value': 5}],
                                  {'group': [(74, 'P')], 'value': 5},
                                  {'group': [(76, 'V')], 'value': 30},
                                  [{'group': [(82, 'A')], 'value': 30},
                                   {'group': [(82, 'C')], 'value': 15},
                                   {'group': [(82, 'F')], 'value': 30},
                                   {'group': [(82, 'L')], 'value': 10},
                                   {'group': [(82, 'M')], 'value': 25},
                                   {'group': [(82, 'S')], 'value': 30},
                                   {'group': [(82, 'T')], 'value': 30}],
                                  [{'group': [(84, 'A')], 'value': 60},
                                   {'group': [(84, 'C')], 'value': 30},
                                   {'group': [(84, 'V')], 'value': 30}],
                                  {'group': [(90, 'M')], 'value': 15},
                                  {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                                  {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                  {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                  {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                                  {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                                  {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                  {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                                  {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                  {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                  {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                                  {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                  {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 5},
                                  {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                                  {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                                  {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 5}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'nelfinavir': ([{'group': [(10, 'F')], 'value': 15},
                                 {'group': [(20, 'T')], 'value': 15},
                                 {'group': [(23, 'I')], 'value': 15},
                                 [{'group': [(24, 'F')], 'value': 10},
                                  {'group': [(24, 'I')], 'value': 10},
                                  {'group': [(24, 'M')], 'value': 10}],
                                 {'group': [(30, 'N')], 'value': 60},
                                 {'group': [(32, 'I')], 'value': 15},
                                 {'group': [(33, 'F')], 'value': 10},
                                 {'group': [(43, 'T')], 'value': 10},
                                 [{'group': [(46, 'I')], 'value': 30},
                                  {'group': [(46, 'L')], 'value': 20},
                                  {'group': [(46, 'V')], 'value': 20}],
                                 [{'group': [(47, 'A')], 'value': 30},
                                  {'group': [(47, 'V')], 'value': 20}],
                                 [{'group': [(48, 'A')], 'value': 30},
                                  {'group': [(48, 'L')], 'value': 30},
                                  {'group': [(48, 'M')], 'value': 30},
                                  {'group': [(48, 'Q')], 'value': 30},
                                  {'group': [(48, 'S')], 'value': 30},
                                  {'group': [(48, 'T')], 'value': 30},
                                  {'group': [(48, 'V')], 'value': 30}],
                                 {'group': [(50, 'V')], 'value': 15},
                                 {'group': [(53, 'L')], 'value': 10},
                                 [{'group': [(54, 'A')], 'value': 20},
                                  {'group': [(54, 'L')], 'value': 20},
                                  {'group': [(54, 'M')], 'value': 20},
                                  {'group': [(54, 'S')], 'value': 20},
                                  {'group': [(54, 'T')], 'value': 20},
                                  {'group': [(54, 'V')], 'value': 20}],
                                 {'group': [(58, 'E')], 'value': 10},
                                 [{'group': [(73, 'A')], 'value': 15},
                                  {'group': [(73, 'C')], 'value': 15},
                                  {'group': [(73, 'D')], 'value': 10},
                                  {'group': [(73, 'S')], 'value': 15},
                                  {'group': [(73, 'T')], 'value': 15},
                                  {'group': [(73, 'V')], 'value': 10}],
                                 {'group': [(74, 'P')], 'value': 20},
                                 {'group': [(76, 'V')], 'value': 10},
                                 [{'group': [(82, 'A')], 'value': 30},
                                  {'group': [(82, 'C')], 'value': 30},
                                  {'group': [(82, 'F')], 'value': 30},
                                  {'group': [(82, 'L')], 'value': 10},
                                  {'group': [(82, 'M')], 'value': 30},
                                  {'group': [(82, 'S')], 'value': 30},
                                  {'group': [(82, 'T')], 'value': 30}],
                                 {'group': [(83, 'D')], 'value': 15},
                                 [{'group': [(84, 'A')], 'value': 60},
                                  {'group': [(84, 'C')], 'value': 60},
                                  {'group': [(84, 'V')], 'value': 60}],
                                 [{'group': [(88, 'D')], 'value': 60},
                                  {'group': [(88, 'G')], 'value': 30},
                                  {'group': [(88, 'S')], 'value': 60},
                                  {'group': [(88, 'T')], 'value': 30}],
                                 {'group': [(89, 'V')], 'value': 10},
                                 {'group': [(90, 'M')], 'value': 60},
                                 {'group': [(11, 'IL'), (32, 'I')], 'value': 5},
                                 {'group': [(11, 'IL'), (54, 'LM')], 'value': 5},
                                 {'group': [(32, 'I'), (47, 'AV')], 'value': 5},
                                 {'group': [(32, 'I'), (54, 'LM')], 'value': 5},
                                 {'group': [(32, 'I'), (76, 'V')], 'value': 5},
                                 {'group': [(32, 'I'), (84, 'V')], 'value': 5},
                                 {'group': [(32, 'I'), (89, 'V')], 'value': 5},
                                 {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                 {'group': [(46, 'ILV'), (76, 'V')], 'value': 10},
                                 {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                 {'group': [(46, 'ILV'), (90, 'M')], 'value': 10},
                                 {'group': [(47, 'AV'), (54, 'LM')], 'value': 5},
                                 {'group': [(47, 'AV'), (84, 'V')], 'value': 5},
                                 {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                                 {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                 {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                                 {'group': [(54, 'LM'), (84, 'V')], 'value': 5},
                                 {'group': [(54, 'LM'), (89, 'V')], 'value': 5},
                                 {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                                 {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'nevirapine': ([{'group': [(98, 'G')], 'value': 30},
                                 [{'group': [(100, 'I')], 'value': 60},
                                  {'group': [(100, 'V')], 'value': 30}],
                                 [{'group': [(101, 'E')], 'value': 30},
                                  {'group': [(101, 'H')], 'value': 15},
                                  {'group': [(101, 'P')], 'value': 60}],
                                 [{'group': [(103, 'H')], 'value': 60},
                                  {'group': [(103, 'N')], 'value': 60},
                                  {'group': [(103, 'S')], 'value': 60},
                                  {'group': [(103, 'T')], 'value': 60}],
                                 [{'group': [(106, 'A')], 'value': 60},
                                  {'group': [(106, 'I')], 'value': 10},
                                  {'group': [(106, 'M')], 'value': 60}],
                                 {'group': [(108, 'I')], 'value': 15},
                                 [{'group': [(138, 'G')], 'value': 10},
                                  {'group': [(138, 'K')], 'value': 10},
                                  {'group': [(138, 'Q')], 'value': 10},
                                  {'group': [(138, 'R')], 'value': 10}],
                                 [{'group': [(179, 'D')], 'value': 10},
                                  {'group': [(179, 'E')], 'value': 10},
                                  {'group': [(179, 'F')], 'value': 15},
                                  {'group': [(179, 'L')], 'value': 10}],
                                 [{'group': [(181, 'C')], 'value': 60},
                                  {'group': [(181, 'F')], 'value': 60},
                                  {'group': [(181, 'G')], 'value': 60},
                                  {'group': [(181, 'I')], 'value': 60},
                                  {'group': [(181, 'S')], 'value': 60},
                                  {'group': [(181, 'V')], 'value': 60}],
                                 [{'group': [(188, 'C')], 'value': 60},
                                  {'group': [(188, 'F')], 'value': 60},
                                  {'group': [(188, 'H')], 'value': 60},
                                  {'group': [(188, 'L')], 'value': 60}],
                                 [{'group': [(190, 'A')], 'value': 60},
                                  {'group': [(190, 'C')], 'value': 60},
                                  {'group': [(190, 'E')], 'value': 60},
                                  {'group': [(190, 'Q')], 'value': 60},
                                  {'group': [(190, 'S')], 'value': 60},
                                  {'group': [(190, 'T')], 'value': 60},
                                  {'group': [(190, 'V')], 'value': 60}],
                                 {'group': [(221, 'Y')], 'value': 15},
                                 {'group': [(225, 'H')], 'value': 45},
                                 [{'group': [(227, 'C')], 'value': 45},
                                  {'group': [(227, 'I')], 'value': 30},
                                  {'group': [(227, 'L')], 'value': 30},
                                  {'group': [(227, 'V')], 'value': 30}],
                                 [{'group': [(230, 'I')], 'value': 30},
                                  {'group': [(230, 'L')], 'value': 60}],
                                 [{'group': [(238, 'N')], 'value': 10},
                                  {'group': [(238, 'T')], 'value': 30}],
                                 {'group': [(318, 'F')], 'value': 30},
                                 {'group': [(348, 'I')], 'value': 15},
                                 {'group': [(101, 'E'), (181, 'C')], 'value': 5},
                                 {'group': [(103, 'R'), (179, 'D')], 'value': 20},
                                 {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                                 {'group': [(98, 'G'), (227, 'CL')], 'value': 15}],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}}),
                 'raltegravir': ([{'group': [(51, 'Y')], 'value': 15},
                                  [{'group': [(66, 'A')], 'value': 15},
                                   {'group': [(66, 'I')], 'value': 15},
                                   {'group': [(66, 'K')], 'value': 60}],
                                  [{'group': [(92, 'G')], 'value': 15},
                                   {'group': [(92, 'Q')], 'value': 30},
                                   {'group': [(92, 'V')], 'value': 30}],
                                  {'group': [(95, 'K')], 'value': 10},
                                  {'group': [(97, 'A')], 'value': 10},
                                  {'group': [(118, 'R')], 'value': 60},
                                  [{'group': [(121, 'C')], 'value': 60},
                                   {'group': [(121, 'Y')], 'value': 60}],
                                  [{'group': [(138, 'A')], 'value': 15},
                                   {'group': [(138, 'K')], 'value': 15},
                                   {'group': [(138, 'T')], 'value': 15}],
                                  [{'group': [(140, 'A')], 'value': 30},
                                   {'group': [(140, 'C')], 'value': 30},
                                   {'group': [(140, 'R')], 'value': 30},
                                   {'group': [(140, 'S')], 'value': 30}],
                                  [{'group': [(143, 'A')], 'value': 60},
                                   {'group': [(143, 'C')], 'value': 60},
                                   {'group': [(143, 'G')], 'value': 60},
                                   {'group': [(143, 'H')], 'value': 60},
                                   {'group': [(143, 'K')], 'value': 60},
                                   {'group': [(143, 'R')], 'value': 60},
                                   {'group': [(143, 'S')], 'value': 60}],
                                  {'group': [(147, 'G')], 'value': 10},
                                  [{'group': [(148, 'H')], 'value': 60},
                                   {'group': [(148, 'K')], 'value': 60},
                                   {'group': [(148, 'N')], 'value': 15},
                                   {'group': [(148, 'R')], 'value': 60}],
                                  [{'group': [(151, 'A')], 'value': 15},
                                   {'group': [(151, 'L')], 'value': 30}],
                                  [{'group': [(153, 'F')], 'value': 10},
                                   {'group': [(153, 'Y')], 'value': 10}],
                                  [{'group': [(155, 'H')], 'value': 60},
                                   {'group': [(155, 'S')], 'value': 30},
                                   {'group': [(155, 'T')], 'value': 30}],
                                  {'group': [(157, 'Q')], 'value': 10},
                                  [{'group': [(163, 'K')], 'value': 15},
                                   {'group': [(163, 'R')], 'value': 15}],
                                  {'group': [(230, 'R')], 'value': 20},
                                  {'group': [(232, 'N')], 'value': 10},
                                  {'group': [(263, 'K')], 'value': 25},
                                  {'group': [(118, 'R'), (138, 'AKT')], 'value': 10},
                                  {'group': [(138, 'AKT'), (140, 'ACS')], 'value': 15},
                                  {'group': [(140, 'ACS'), (148, 'HKR'), (149, 'A')],
                                   'value': 10},
                                  {'group': [(74, 'FIM'), (118, 'R')], 'value': 10},
                                  {'group': [(74, 'FIM'), (148, 'HKR')], 'value': 15},
                                  {'group': [(92, 'Q'), (155, 'H')], 'value': 10},
                                  {'group': [(97, 'A'), (118, 'R')], 'value': 10}],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'rilpivirine': ([{'group': [(98, 'G')], 'value': 15},
                                  [{'group': [(100, 'I')], 'value': 60},
                                   {'group': [(100, 'V')], 'value': 15}],
                                  [{'group': [(101, 'E')], 'value': 45},
                                   {'group': [(101, 'H')], 'value': 10},
                                   {'group': [(101, 'P')], 'value': 60}],
                                  {'group': [(106, 'I')], 'value': 10},
                                  [{'group': [(138, 'A')], 'value': 15},
                                   {'group': [(138, 'G')], 'value': 15},
                                   {'group': [(138, 'K')], 'value': 45},
                                   {'group': [(138, 'Q')], 'value': 15},
                                   {'group': [(138, 'R')], 'value': 15}],
                                  [{'group': [(179, 'D')], 'value': 10},
                                   {'group': [(179, 'E')], 'value': 10},
                                   {'group': [(179, 'F')], 'value': 15},
                                   {'group': [(179, 'L')], 'value': 15}],
                                  [{'group': [(181, 'C')], 'value': 45},
                                   {'group': [(181, 'F')], 'value': 30},
                                   {'group': [(181, 'G')], 'value': 30},
                                   {'group': [(181, 'I')], 'value': 60},
                                   {'group': [(181, 'S')], 'value': 30},
                                   {'group': [(181, 'V')], 'value': 60}],
                                  [{'group': [(188, 'F')], 'value': 30},
                                   {'group': [(188, 'L')], 'value': 60}],
                                  [{'group': [(190, 'A')], 'value': 15},
                                   {'group': [(190, 'C')], 'value': 10},
                                   {'group': [(190, 'E')], 'value': 60},
                                   {'group': [(190, 'Q')], 'value': 45},
                                   {'group': [(190, 'S')], 'value': 15},
                                   {'group': [(190, 'T')], 'value': 10},
                                   {'group': [(190, 'V')], 'value': 10}],
                                  {'group': [(221, 'Y')], 'value': 15},
                                  {'group': [(227, 'C')], 'value': 45},
                                  [{'group': [(230, 'I')], 'value': 30},
                                   {'group': [(230, 'L')], 'value': 60}],
                                  {'group': [(101, 'E'), (184, 'I')], 'value': 15},
                                  {'group': [(103, 'R'), (179, 'D')], 'value': 15},
                                  {'group': [(106, 'I'), (181, 'C')], 'value': 10},
                                  {'group': [(138, 'K'), (184, 'I')], 'value': 15},
                                  {'group': [(181, 'CIV'), (190, 'ACSTV')], 'value': 10},
                                  {'group': [(181, 'CIV'), (221, 'Y')], 'value': 10},
                                  {'group': [(98, 'G'), (181, 'C')], 'value': 5},
                                  {'group': [(98, 'G'), (227, 'CL')], 'value': 15},
                                  [{'group': [(179, 'F'), (181, 'C')], 'value': 15},
                                   {'group': [(179, 'T'), (181, 'C')], 'value': 10}]],
                                 {1: {'max': 9, 'min': -inf},
                                  2: {'max': 14, 'min': 10},
                                  3: {'max': 29, 'min': 15},
                                  4: {'max': 59, 'min': 30},
                                  5: {'max': inf, 'min': 60}}),
                 'saquinavir/r': ([{'group': [(20, 'T')], 'value': 5},
                                   [{'group': [(24, 'F')], 'value': 5},
                                    {'group': [(24, 'I')], 'value': 10},
                                    {'group': [(24, 'M')], 'value': 5}],
                                   {'group': [(33, 'F')], 'value': 5},
                                   [{'group': [(46, 'I')], 'value': 10},
                                    {'group': [(46, 'L')], 'value': 10},
                                    {'group': [(46, 'V')], 'value': 5}],
                                   [{'group': [(48, 'A')], 'value': 60},
                                    {'group': [(48, 'L')], 'value': 60},
                                    {'group': [(48, 'M')], 'value': 60},
                                    {'group': [(48, 'Q')], 'value': 60},
                                    {'group': [(48, 'S')], 'value': 60},
                                    {'group': [(48, 'T')], 'value': 60},
                                    {'group': [(48, 'V')], 'value': 60}],
                                   [{'group': [(50, 'L')], 'value': -5},
                                    {'group': [(50, 'V')], 'value': 15}],
                                   {'group': [(53, 'L')], 'value': 15},
                                   [{'group': [(54, 'A')], 'value': 15},
                                    {'group': [(54, 'L')], 'value': 15},
                                    {'group': [(54, 'M')], 'value': 15},
                                    {'group': [(54, 'S')], 'value': 15},
                                    {'group': [(54, 'T')], 'value': 15},
                                    {'group': [(54, 'V')], 'value': 15}],
                                   [{'group': [(73, 'A')], 'value': 15},
                                    {'group': [(73, 'C')], 'value': 15},
                                    {'group': [(73, 'D')], 'value': 10},
                                    {'group': [(73, 'S')], 'value': 15},
                                    {'group': [(73, 'T')], 'value': 15},
                                    {'group': [(73, 'V')], 'value': 10}],
                                   {'group': [(74, 'P')], 'value': 10},
                                   [{'group': [(82, 'A')], 'value': 15},
                                    {'group': [(82, 'C')], 'value': 15},
                                    {'group': [(82, 'F')], 'value': 10},
                                    {'group': [(82, 'L')], 'value': 10},
                                    {'group': [(82, 'M')], 'value': 15},
                                    {'group': [(82, 'S')], 'value': 15},
                                    {'group': [(82, 'T')], 'value': 15}],
                                   {'group': [(83, 'D')], 'value': 10},
                                   [{'group': [(84, 'A')], 'value': 60},
                                    {'group': [(84, 'C')], 'value': 60},
                                    {'group': [(84, 'V')], 'value': 60}],
                                   [{'group': [(88, 'D')], 'value': 10},
                                    {'group': [(88, 'S')], 'value': 15}],
                                   {'group': [(90, 'M')], 'value': 45},
                                   {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5},
                                   {'group': [(46, 'ILV'), (82, 'ACFLMST')], 'value': 10},
                                   {'group': [(46, 'ILV'), (90, 'M')], 'value': 5},
                                   {'group': [(53, 'L'), (90, 'M')], 'value': 10},
                                   {'group': [(54, 'ALMSTV'), (82, 'ACFLMST')], 'value': 10},
                                   {'group': [(54, 'ALMSTV'), (90, 'M')], 'value': 10},
                                   {'group': [(73, 'ACSTV'), (90, 'M')], 'value': 10},
                                   {'group': [(82, 'ACFLMST'), (90, 'M')], 'value': 10}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'stavudine': ([{'group': [(41, 'L')], 'value': 15},
                                {'group': [(62, 'V')], 'value': 5},
                                [{'group': [(65, 'E')], 'value': 10},
                                 {'group': [(65, 'N')], 'value': 30},
                                 {'group': [(65, 'R')], 'value': 60}],
                                [{'group': [(67, 'E')], 'value': 10},
                                 {'group': [(67, 'G')], 'value': 10},
                                 {'group': [(67, 'H')], 'value': 10},
                                 {'group': [(67, 'N')], 'value': 15},
                                 {'group': [(67, 'S')], 'value': 10},
                                 {'group': [(67, 'T')], 'value': 10},
                                 {'group': [(67, 'd')], 'value': 30}],
                                {'group': [(68, 'd')], 'value': 30},
                                [{'group': [(69, 'G')], 'value': 10},
                                 {'group': [(69, 'i')], 'value': 60},
                                 {'group': [(69, 'd')], 'value': 30},
                                 {'group': [(69, 'D')], 'value': 10}],
                                [{'group': [(70, 'E')], 'value': 15},
                                 {'group': [(70, 'G')], 'value': 15},
                                 {'group': [(70, 'N')], 'value': 15},
                                 {'group': [(70, 'Q')], 'value': 15},
                                 {'group': [(70, 'R')], 'value': 15},
                                 {'group': [(70, 'S')], 'value': 15},
                                 {'group': [(70, 'T')], 'value': 15},
                                 {'group': [(70, 'd')], 'value': 30}],
                                [{'group': [(75, 'I')], 'value': 5},
                                 {'group': [(75, 'M')], 'value': 30},
                                 {'group': [(75, 'A')], 'value': 30},
                                 {'group': [(75, 'S')], 'value': 30},
                                 {'group': [(75, 'T')], 'value': 60}],
                                {'group': [(77, 'L')], 'value': 10},
                                {'group': [(116, 'Y')], 'value': 10},
                                [{'group': [(151, 'L')], 'value': 30},
                                 {'group': [(151, 'M')], 'value': 60}],
                                [{'group': [(184, 'I')], 'value': -10},
                                 {'group': [(184, 'V')], 'value': -10}],
                                {'group': [(210, 'W')], 'value': 15},
                                [{'group': [(215, 'A')], 'value': 20},
                                 {'group': [(215, 'C')], 'value': 20},
                                 {'group': [(215, 'D')], 'value': 20},
                                 {'group': [(215, 'E')], 'value': 20},
                                 {'group': [(215, 'F')], 'value': 40},
                                 {'group': [(215, 'I')], 'value': 20},
                                 {'group': [(215, 'L')], 'value': 20},
                                 {'group': [(215, 'N')], 'value': 20},
                                 {'group': [(215, 'S')], 'value': 20},
                                 {'group': [(215, 'V')], 'value': 20},
                                 {'group': [(215, 'Y')], 'value': 40}],
                                [{'group': [(219, 'E')], 'value': 10},
                                 {'group': [(219, 'N')], 'value': 10},
                                 {'group': [(219, 'Q')], 'value': 10},
                                 {'group': [(219, 'R')], 'value': 10},
                                 {'group': [(219, 'W')], 'value': 10}],
                                {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                                 'value': 5},
                                {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                                {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                                 'value': 5},
                                {'group': [(41, 'L'), (67, 'EGN'), (215, 'FY')], 'value': 5},
                                {'group': [(67, 'EGN'), (215, 'FY'), (219, 'ENQR')],
                                 'value': 5},
                                {'group': [(67, 'EGN'), (70, 'R'), (219, 'ENQR')], 'value': 10},
                                {'group': [(70, 'R'), (215, 'FY')], 'value': 5},
                                {'group': [(77, 'L'), (116, 'Y'), (151, 'M')], 'value': 10},
                                [{'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 5},
                                 {'group': [(210, 'W'), (215, 'FY')], 'value': 10}],
                                [{'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 5},
                                 {'group': [(41, 'L'), (215, 'FY')], 'value': 10}],
                                {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                                {'group': [(70, 'EGNQST'), (184, 'IV')], 'value': 10}],
                               {1: {'max': 9, 'min': -inf},
                                2: {'max': 14, 'min': 10},
                                3: {'max': 29, 'min': 15},
                                4: {'max': 59, 'min': 30},
                                5: {'max': inf, 'min': 60}}),
                 'tenofovir': ([{'group': [(41, 'L')], 'value': 5},
                                {'group': [(62, 'V')], 'value': 5},
                                [{'group': [(65, 'E')], 'value': 10},
                                 {'group': [(65, 'N')], 'value': 45},
                                 {'group': [(65, 'R')], 'value': 50}],
                                [{'group': [(67, 'E')], 'value': 5},
                                 {'group': [(67, 'G')], 'value': 5},
                                 {'group': [(67, 'H')], 'value': 5},
                                 {'group': [(67, 'N')], 'value': 5},
                                 {'group': [(67, 'S')], 'value': 5},
                                 {'group': [(67, 'T')], 'value': 5},
                                 {'group': [(67, 'd')], 'value': 30}],
                                {'group': [(68, 'd')], 'value': 15},
                                [{'group': [(69, 'G')], 'value': 5},
                                 {'group': [(69, 'i')], 'value': 60},
                                 {'group': [(69, 'd')], 'value': 15}],
                                [{'group': [(70, 'E')], 'value': 15},
                                 {'group': [(70, 'G')], 'value': 15},
                                 {'group': [(70, 'N')], 'value': 15},
                                 {'group': [(70, 'Q')], 'value': 15},
                                 {'group': [(70, 'R')], 'value': 5},
                                 {'group': [(70, 'S')], 'value': 15},
                                 {'group': [(70, 'T')], 'value': 15},
                                 {'group': [(70, 'd')], 'value': 15}],
                                {'group': [(74, 'I')], 'value': 5},
                                {'group': [(75, 'I')], 'value': 5},
                                {'group': [(77, 'L')], 'value': 5},
                                {'group': [(115, 'F')], 'value': 15},
                                {'group': [(116, 'Y')], 'value': 5},
                                [{'group': [(151, 'L')], 'value': 10},
                                 {'group': [(151, 'M')], 'value': 15}],
                                [{'group': [(184, 'I')], 'value': -10},
                                 {'group': [(184, 'V')], 'value': -10}],
                                {'group': [(210, 'W')], 'value': 5},
                                [{'group': [(215, 'F')], 'value': 10},
                                 {'group': [(215, 'I')], 'value': 5},
                                 {'group': [(215, 'V')], 'value': 5},
                                 {'group': [(215, 'Y')], 'value': 10}],
                                [{'group': [(219, 'E')], 'value': 5},
                                 {'group': [(219, 'N')], 'value': 5},
                                 {'group': [(219, 'Q')], 'value': 5},
                                 {'group': [(219, 'R')], 'value': 5},
                                 {'group': [(219, 'W')], 'value': 5}],
                                {'group': [(151, 'M'), (184, 'IV')], 'value': 15},
                                {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                                {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                                {'group': [(115, 'F'), (184, 'IV')], 'value': 5},
                                {'group': [(70, 'EGNQST'), (184, 'IV')], 'value': 10},
                                {'group': [(65, 'R'), (68, 'NGR')], 'value': 5},
                                {'group': [(62, 'V'), (65, 'R')], 'value': 5},
                                {'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                                {'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                                {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                                {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 10},
                                {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                                 'value': 5},
                                {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                                 'value': 5},
                                {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                                {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')],
                                 'value': 5},
                                {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')],
                                 'value': 10}],
                               {1: {'max': 9, 'min': -inf},
                                2: {'max': 14, 'min': 10},
                                3: {'max': 29, 'min': 15},
                                4: {'max': 59, 'min': 30},
                                5: {'max': inf, 'min': 60}}),
                 'tipranavir/r': ([{'group': [(24, 'I')], 'value': -5},
                                   {'group': [(32, 'I')], 'value': 5},
                                   {'group': [(33, 'F')], 'value': 10},
                                   {'group': [(43, 'T')], 'value': 10},
                                   [{'group': [(46, 'I')], 'value': 5},
                                    {'group': [(46, 'L')], 'value': 10},
                                    {'group': [(46, 'V')], 'value': 5}],
                                   [{'group': [(47, 'A')], 'value': 30},
                                    {'group': [(47, 'V')], 'value': 30}],
                                   [{'group': [(50, 'L')], 'value': -5},
                                    {'group': [(50, 'V')], 'value': -5}],
                                   [{'group': [(54, 'A')], 'value': 20},
                                    {'group': [(54, 'L')], 'value': -10},
                                    {'group': [(54, 'M')], 'value': 20},
                                    {'group': [(54, 'S')], 'value': 20},
                                    {'group': [(54, 'T')], 'value': 20},
                                    {'group': [(54, 'V')], 'value': 20}],
                                   {'group': [(58, 'E')], 'value': 15},
                                   {'group': [(74, 'P')], 'value': 25},
                                   {'group': [(76, 'V')], 'value': -5},
                                   [{'group': [(82, 'C')], 'value': 10},
                                    {'group': [(82, 'L')], 'value': 45},
                                    {'group': [(82, 'M')], 'value': 10},
                                    {'group': [(82, 'S')], 'value': 30},
                                    {'group': [(82, 'T')], 'value': 45}],
                                   {'group': [(83, 'D')], 'value': 25},
                                   [{'group': [(84, 'A')], 'value': 60},
                                    {'group': [(84, 'C')], 'value': 30},
                                    {'group': [(84, 'V')], 'value': 30}],
                                   {'group': [(46, 'IL'), (84, 'V'), (90, 'M')], 'value': 5}],
                                  {1: {'max': 9, 'min': -inf},
                                   2: {'max': 14, 'min': 10},
                                   3: {'max': 29, 'min': 15},
                                   4: {'max': 59, 'min': 30},
                                   5: {'max': inf, 'min': 60}}),
                 'zidovudine': ([{'group': [(41, 'L')], 'value': 15},
                                 {'group': [(62, 'V')], 'value': 5},
                                 {'group': [(65, 'R')], 'value': -10},
                                 [{'group': [(67, 'E')], 'value': 15},
                                  {'group': [(67, 'G')], 'value': 15},
                                  {'group': [(67, 'H')], 'value': 15},
                                  {'group': [(67, 'N')], 'value': 15},
                                  {'group': [(67, 'S')], 'value': 15},
                                  {'group': [(67, 'T')], 'value': 15},
                                  {'group': [(67, 'd')], 'value': 30}],
                                 [{'group': [(69, 'G')], 'value': 5},
                                  {'group': [(69, 'i')], 'value': 60}],
                                 {'group': [(70, 'R')], 'value': 30},
                                 [{'group': [(75, 'I')], 'value': 5},
                                  {'group': [(75, 'M')], 'value': 10},
                                  {'group': [(75, 'A')], 'value': 10},
                                  {'group': [(75, 'S')], 'value': 10},
                                  {'group': [(75, 'T')], 'value': 10}],
                                 {'group': [(77, 'L')], 'value': 10},
                                 {'group': [(116, 'Y')], 'value': 10},
                                 [{'group': [(151, 'L')], 'value': 30},
                                  {'group': [(151, 'M')], 'value': 60}],
                                 [{'group': [(184, 'I')], 'value': -10},
                                  {'group': [(184, 'V')], 'value': -10}],
                                 {'group': [(210, 'W')], 'value': 15},
                                 [{'group': [(215, 'A')], 'value': 10},
                                  {'group': [(215, 'C')], 'value': 10},
                                  {'group': [(215, 'D')], 'value': 10},
                                  {'group': [(215, 'E')], 'value': 10},
                                  {'group': [(215, 'F')], 'value': 60},
                                  {'group': [(215, 'I')], 'value': 20},
                                  {'group': [(215, 'L')], 'value': 10},
                                  {'group': [(215, 'N')], 'value': 10},
                                  {'group': [(215, 'S')], 'value': 10},
                                  {'group': [(215, 'V')], 'value': 20},
                                  {'group': [(215, 'Y')], 'value': 60}],
                                 [{'group': [(219, 'E')], 'value': 10},
                                  {'group': [(219, 'N')], 'value': 10},
                                  {'group': [(219, 'Q')], 'value': 10},
                                  {'group': [(219, 'R')], 'value': 10},
                                  {'group': [(219, 'W')], 'value': 10}],
                                 {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                                 {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                                 {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 10},
                                 [{'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                                  {'group': [(41, 'L'), (215, 'ACDEILNSV')], 'value': 10}],
                                 [{'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                                  {'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 10}],
                                 {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                                 {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')],
                                  'value': 5},
                                 {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')],
                                  'value': 5},
                                 {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')],
                                  'value': 5},
                                 {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')],
                                  'value': 5},
                                 {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')],
                                  'value': 15}],
                                {1: {'max': 9, 'min': -inf},
                                 2: {'max': 14, 'min': 10},
                                 3: {'max': 29, 'min': 15},
                                 4: {'max': 59, 'min': 30},
                                 5: {'max': inf, 'min': 60}})}
               
        res_drugs = self.algorithm.parse_drugs(self.root)

        self.assertEqual(exp_drugs, res_drugs)

    def testParseCondition(self):
        # Setting params
        condition = self.root.find('.//CONDITION').text
        exp_drms = \
                [{'group': [(41, 'L')], 'value': 5},
                 {'group': [(62, 'V')], 'value': 5},
                 [{'group': [(65, 'E')], 'value': 10},
                  {'group': [(65, 'N')], 'value': 30},
                  {'group': [(65, 'R')], 'value': 45}],
                 [{'group': [(67, 'E')], 'value': 5},
                  {'group': [(67, 'G')], 'value': 5},
                  {'group': [(67, 'H')], 'value': 5},
                  {'group': [(67, 'N')], 'value': 5},
                  {'group': [(67, 'S')], 'value': 5},
                  {'group': [(67, 'T')], 'value': 5},
                  {'group': [(67, 'd')], 'value': 30}],
                 {'group': [(68, 'd')], 'value': 15},
                 [{'group': [(69, 'G')], 'value': 10},
                  {'group': [(69, 'i')], 'value': 60},
                  {'group': [(69, 'd')], 'value': 15}],
                 [{'group': [(70, 'E')], 'value': 15},
                  {'group': [(70, 'G')], 'value': 15},
                  {'group': [(70, 'N')], 'value': 15},
                  {'group': [(70, 'Q')], 'value': 15},
                  {'group': [(70, 'R')], 'value': 5},
                  {'group': [(70, 'S')], 'value': 15},
                  {'group': [(70, 'T')], 'value': 15},
                  {'group': [(70, 'd')], 'value': 15}],
                 [{'group': [(74, 'I')], 'value': 15}, {'group': [(74, 'V')], 'value': 30}],
                 {'group': [(75, 'I')], 'value': 5},
                 {'group': [(77, 'L')], 'value': 5},
                 {'group': [(115, 'F')], 'value': 30},
                 {'group': [(116, 'Y')], 'value': 5},
                 [{'group': [(151, 'L')], 'value': 30}, {'group': [(151, 'M')], 'value': 60}],
                 [{'group': [(184, 'I')], 'value': 15}, {'group': [(184, 'V')], 'value': 15}],
                 {'group': [(210, 'W')], 'value': 5},
                 [{'group': [(215, 'F')], 'value': 10},
                  {'group': [(215, 'I')], 'value': 5},
                  {'group': [(215, 'V')], 'value': 5},
                  {'group': [(215, 'Y')], 'value': 10}],
                 [{'group': [(219, 'E')], 'value': 5},
                  {'group': [(219, 'N')], 'value': 5},
                  {'group': [(219, 'Q')], 'value': 5},
                  {'group': [(219, 'R')], 'value': 5},
                  {'group': [(219, 'W')], 'value': 5}],
                 {'group': [(151, 'M'), (184, 'IV')], 'value': 10},
                 {'group': [(65, 'RN'), (151, 'M')], 'value': 10},
                 {'group': [(77, 'L'), (116, 'Y'), (151, 'ML')], 'value': 15},
                 {'group': [(74, 'V'), (184, 'IV')], 'value': 15},
                 {'group': [(115, 'F'), (184, 'IV')], 'value': 15},
                 {'group': [(41, 'L'), (215, 'FY')], 'value': 10},
                 {'group': [(210, 'W'), (215, 'FY')], 'value': 10},
                 {'group': [(41, 'L'), (210, 'W')], 'value': 10},
                 {'group': [(41, 'L'), (210, 'W'), (215, 'FY')], 'value': 10},
                 {'group': [(41, 'L'), (44, 'AD'), (210, 'W'), (215, 'FY')], 'value': 5},
                 {'group': [(40, 'F'), (41, 'L'), (210, 'W'), (215, 'FY')], 'value': 5},
                 {'group': [(41, 'L'), (67, 'EGNHST'), (215, 'FY')], 'value': 5},
                 {'group': [(67, 'EGNHST'), (215, 'FY'), (219, 'ENQRW')], 'value': 5},
                 {'group': [(67, 'EGNHST'), (70, 'R'), (219, 'ENQRW')], 'value': 10},
                 {'group': [(41, 'L'), (184, 'VI'), (215, 'FY')], 'value': 10},
                 {'group': [(67, 'EGNHST'), (70, 'R'), (184, 'VI'), (219, 'ENQRW')],
                  'value': 10}]               
        
        res_drms = self.algorithm.parse_condition(condition)

        self.assertEqual(exp_drms, res_drms)

    def testParseScores(self):
        # Setting params
        drm_lib = []
        drm = 'MAX ((210W AND 215ACDEILNSV) => 5, (210W AND 215FY) => 10)'
        chunk = '(210W AND 215ACDEILNSV) => 5'
        iter = 0

        exp_drm_lib = [{'group': [(210, 'W'), (215, 'ACDEILNSV')], 'value': 5}]
        self.algorithm._parse_scores(drm_lib, drm, chunk, iter)
        self.assertEqual(exp_drm_lib, drm_lib)

        # if iter parameter is set to anything greater than 1
        with self.assertRaises(IndexError) as err: 
            self.algorithm._parse_scores(drm_lib, drm, chunk, 3)

        with self.assertRaises(IndexError) as err: 
            self.algorithm._parse_scores(drm_lib, drm, chunk, 2)

        # Setting params
        drm_lib = []
        drm = '(67EGN AND 70R AND 184IV AND 219ENQR) => 20'
        chunk = '(67EGN AND 70R AND 184IV AND 219ENQR) => 20'
        iter = 0

        exp_drm_lib = [{'group': [(67, 'EGN'), (70, 'R'), (184, 'IV'), (219, 'ENQR')], 'value': 20}]
        self.algorithm._parse_scores(drm_lib, drm, chunk, iter)
        self.assertEqual(exp_drm_lib, drm_lib)

        # if iter parameter is set to anything greater than 0
        with self.assertRaises(IndexError) as err: 
            self.algorithm._parse_scores(drm_lib, drm, chunk, 1)

        with self.assertRaises(IndexError) as err: 
            self.algorithm._parse_scores(drm_lib, drm, chunk, 2)

    def testParseComments(self):
        exp_comments = \
                {'CA': {'105TSE': 'CA105TSE',
                        '107NS': 'CA107NS',
                        '56I': 'CA56I',
                        '57S': 'CA57S',
                        '66I': 'CA66I',
                        '67H': 'CA67H',
                        '70NHRS': 'CA70NHRS',
                        '74DS': 'CA74DS'},
                 'IN': {'118ACDEFHIKLMNPQSTVWYid': 'IN118ACDEFHIKLMNPQSTVWYid',
                        '118R': 'IN118R',
                        '119R': 'IN119R',
                        '121ADEGHIKLMNPQRSTVWid': 'IN121ADEGHIKLMNPQRSTVWid',
                        '121C': 'IN121C',
                        '121Y': 'IN121Y',
                        '128T': 'IN128T',
                        '138CFGHILMNPQRSVWYid': 'IN138CFGHILMNPQRSVWYid',
                        '138D': 'IN138D',
                        '138KAT': 'IN138KAT',
                        '140DEFHIKLMNPQTVWYid': 'IN140DEFHIKLMNPQTVWYid',
                        '140R': 'IN140R',
                        '140SAC': 'IN140SAC',
                        '143CRH': 'IN143CRH',
                        '143DEFILMNPQTVWid': 'IN143DEFILMNPQTVWid',
                        '143KGSA': 'IN143KGSA',
                        '145ACDEFGHIKLMNQRTVWYid': 'IN145ACDEFGHIKLMNQRTVWYid',
                        '145S': 'IN145S',
                        '146ACDEFGHIKLMNRSTVWYid': 'IN146ACDEFGHIKLMNRSTVWYid',
                        '146PRL': 'IN146PRL',
                        '147ACDEFHIKLMNPQRTVWYid': 'IN147ACDEFHIKLMNPQRTVWYid',
                        '147G': 'IN147G',
                        '148ACDEFGILMPSTVWYid': 'IN148ACDEFGILMPSTVWYid',
                        '148HKR': 'IN148HKR',
                        '148N': 'IN148N',
                        '149A': 'IN149A',
                        '151A': 'IN151A',
                        '151CDEFGHKMNPQRSTWYid': 'IN151CDEFGHKMNPQRSTWYid',
                        '151I': 'IN151I',
                        '151L': 'IN151L',
                        '153ACDEGHIKLMNPQRTVWid': 'IN153ACDEGHIKLMNPQRTVWid',
                        '153YFA': 'IN153YFA',
                        '155ACDEFGIKLMPQRVWYid': 'IN155ACDEFGIKLMPQRVWYid',
                        '155H': 'IN155H',
                        '155ST': 'IN155ST',
                        '157Q': 'IN157Q',
                        '163RK': 'IN163RK',
                        '230N': 'IN230N',
                        '230R': 'IN230R',
                        '232N': 'IN232N',
                        '263ACDEFGHILMNPQSTVWYid': 'IN263ACDEFGHILMNPQSTVWYid',
                        '263K': 'IN263K',
                        '49G': 'IN49G',
                        '50I': 'IN50I',
                        '51ACDEFGIKLMNPQRSTVWid': 'IN51ACDEFGIKLMNPQRSTVWid',
                        '51Y': 'IN51Y',
                        '66AI': 'IN66AI',
                        '66CDEFGHLMNPQRSVWYid': 'IN66CDEFGHLMNPQRSVWYid',
                        '66K': 'IN66K',
                        '74F': 'IN74F',
                        '74I': 'IN74I',
                        '74M': 'IN74M',
                        '92ACDFHIKLMNPRSTWYid': 'IN92ACDFHIKLMNPRSTWYid',
                        '92GV': 'IN92GV',
                        '92Q': 'IN92Q',
                        '95K': 'IN95K',
                        '97A': 'IN97A'},
                 'PR': {'10ACDEGHKMNPQSTWid': 'PR10ACDEGHKMNPQSTWid',
                        '10F': 'PR10F',
                        '10IV': 'PR10IV',
                        '10RY': 'PR10RY',
                        '11IL': 'PR11IL',
                        '20I': 'PR20I',
                        '20MV': 'PR20MV',
                        '20R': 'PR20R',
                        '20T': 'PR20T',
                        '23I': 'PR23I',
                        '24ACDEGHKNPQRSTVWYid': 'PR24ACDEGHKNPQRSTVWYid',
                        '24FM': 'PR24FM',
                        '24I': 'PR24I',
                        '30ACEFGHIKLMPQRSTVWYid': 'PR30ACEFGHIKLMPQRSTVWYid',
                        '30N': 'PR30N',
                        '32ACDEFGHKLMNPQRSTWYid': 'PR32ACDEFGHKLMNPQRSTWYid',
                        '32I': 'PR32I',
                        '33F': 'PR33F',
                        '33IV': 'PR33IV',
                        '33i OR 34i OR 35i OR 36i OR 37i OR 38i OR 39i OR 40i OR 41i': 'PR33-41i',
                        '43T': 'PR43T',
                        '46ACDEFGHKNPQRSTWYid': 'PR46ACDEFGHKNPQRSTWYid',
                        '46IL': 'PR46IL',
                        '46V': 'PR46V',
                        '47A': 'PR47A',
                        '47CDEFGHKLMNPQRSTWYid': 'PR47CDEFGHKLMNPQRSTWYid',
                        '47V': 'PR47V',
                        '48ASTQL': 'PR48ASTQL',
                        '48CDEFHIKNPRWYid': 'PR48CDEFHIKNPRWYid',
                        '48M': 'PR48M',
                        '48V': 'PR48V',
                        '50ACDEFGHKMNPQRSTWYid': 'PR50ACDEFGHKMNPQRSTWYid',
                        '50L': 'PR50L',
                        '50V': 'PR50V',
                        '53LY': 'PR53LY',
                        '54ATS': 'PR54ATS',
                        '54CDEFGHKNPQRWYid': 'PR54CDEFGHKNPQRWYid',
                        '54LM': 'PR54LM',
                        '54V': 'PR54V',
                        '58E': 'PR58E',
                        '71IL': 'PR71IL',
                        '71TV': 'PR71TV',
                        '73EFHIKLMNPQRWYid': 'PR73EFHIKLMNPQRWYid',
                        '73STCA': 'PR73STCA',
                        '74P': 'PR74P',
                        '74S': 'PR74S',
                        '76ACDEFGHIKMNPQRSTWYid': 'PR76ACDEFGHIKMNPQRSTWYid',
                        '76V': 'PR76V',
                        '82A': 'PR82A',
                        '82C': 'PR82C',
                        '82DEGHKNPQRWYid': 'PR82DEGHKNPQRWYid',
                        '82F': 'PR82F',
                        '82I': 'PR82I',
                        '82L': 'PR82L',
                        '82M': 'PR82M',
                        '82TS': 'PR82TS',
                        '83D': 'PR83D',
                        '84AC': 'PR84AC',
                        '84DEFGHKLMNPQRSTWYid': 'PR84DEFGHKLMNPQRSTWYid',
                        '84V': 'PR84V',
                        '85V': 'PR85V',
                        '88ACEFHIKLMPQRVWYid': 'PR88ACEFHIKLMPQRVWYid',
                        '88D': 'PR88D',
                        '88S': 'PR88S',
                        '88TG': 'PR88TG',
                        '89VT': 'PR89VT',
                        '90ACDEFGHIKNPQRSTVWYid': 'PR90ACDEFGHIKNPQRSTVWYid',
                        '90M': 'PR90M'},
                 'RT': {'100ACDEFGHKMNPQRSTWYid': 'RT100ACDEFGHKMNPQRSTWYid',
                        '100I': 'RT100I',
                        '100V': 'RT100V',
                        '101CDFGILMSVWYid': 'RT101CDFGILMSVWYid',
                        '101E': 'RT101E',
                        '101H': 'RT101H',
                        '101NAT': 'RT101NAT',
                        '101P': 'RT101P',
                        '101Q': 'RT101Q',
                        '103ACDFGILMPVWYid': 'RT103ACDFGILMPVWYid',
                        '103EQ': 'RT103EQ',
                        '103H': 'RT103H',
                        '103N': 'RT103N',
                        '103R': 'RT103R',
                        '103S': 'RT103S',
                        '103T': 'RT103T',
                        '106A': 'RT106A',
                        '106CDEFGHKLNPQRSTWYid': 'RT106CDEFGHKLNPQRSTWYid',
                        '106I': 'RT106I',
                        '106M': 'RT106M',
                        '108ACDEFGHKLMNPQRSTWYid': 'RT108ACDEFGHKLMNPQRSTWYid',
                        '108I': 'RT108I',
                        '115ACDEGHIKLMNPQRSTVWid': 'RT115ACDEGHIKLMNPQRSTVWid',
                        '115F': 'RT115F',
                        '116ACDEGHIKLMNPQRSTVWid': 'RT116ACDEGHIKLMNPQRSTVWid',
                        '116Y': 'RT116Y',
                        '118I': 'RT118I',
                        '132ML': 'RT132ML',
                        '138A': 'RT138A',
                        '138CDFHILMNPSTVWYid': 'RT138CDFHILMNPSTVWYid',
                        '138K': 'RT138K',
                        '138QG': 'RT138QG',
                        '138R': 'RT138R',
                        '151ACDEFGHIKNPRSTVWYid': 'RT151ACDEFGHIKNPRSTVWYid',
                        '151L': 'RT151L',
                        '151M': 'RT151M',
                        '179ACGHKMNPQRSWYid': 'RT179ACGHKMNPQRSWYid',
                        '179DE': 'RT179DE',
                        '179F': 'RT179F',
                        '179I': 'RT179I',
                        '179L': 'RT179L',
                        '179T': 'RT179T',
                        '181ADEHKLMNPQRTWid': 'RT181ADEHKLMNPQRTWid',
                        '181C': 'RT181C',
                        '181FSG': 'RT181FSG',
                        '181IV': 'RT181IV',
                        '184ACDEFGHKLNPQRSTWYid': 'RT184ACDEFGHKLNPQRSTWYid',
                        '184VI': 'RT184VI',
                        '188ADEGIKMNPQRSTVWid': 'RT188ADEGIKMNPQRSTVWid',
                        '188C': 'RT188C',
                        '188F': 'RT188F',
                        '188H': 'RT188H',
                        '188L': 'RT188L',
                        '190A': 'RT190A',
                        '190CTV': 'RT190CTV',
                        '190DFHIKLMNPWYid': 'RT190DFHIKLMNPWYid',
                        '190EQ': 'RT190EQ',
                        '190S': 'RT190S',
                        '210W': 'RT210W',
                        '215GHKMPQRWid': 'RT215GHKMPQRWid',
                        '215SCDEIVALN': 'RT215SCDEIVALN',
                        '215YF': 'RT215YF',
                        '219ACDFGHILMPSTVYid': 'RT219ACDFGHILMPSTVYid',
                        '219EQNR': 'RT219EQNR',
                        '219W': 'RT219W',
                        '221Y': 'RT221Y',
                        '225ACDEFGIKLMNQRSTVWYid': 'RT225ACDEFGIKLMNQRSTVWYid',
                        '225H': 'RT225H',
                        '227ADEGHKMNPQRSTWYid': 'RT227ADEGHKMNPQRSTWYid',
                        '227C': 'RT227C',
                        '227ILV': 'RT227ILV',
                        '230ACDEFGHKNPQRSTVWYid': 'RT230ACDEFGHKNPQRSTVWYid',
                        '230I': 'RT230I',
                        '230L': 'RT230L',
                        '234ACDEFGHKMNPQRSTWYid': 'RT234ACDEFGHKMNPQRSTWYid',
                        '234I': 'RT234I',
                        '236ACDEFGHIKMNQRSTVWYid': 'RT236ACDEFGHIKMNQRSTVWYid',
                        '236L': 'RT236L',
                        '238ACDEFGHILMPQSVWYid': 'RT238ACDEFGHILMPQSVWYid',
                        '238TN': 'RT238TN',
                        '318ACDEGHIKLMNPQRSTVWid': 'RT318ACDEGHIKLMNPQRSTVWid',
                        '318F': 'RT318F',
                        '348ACDEFGHKLMPQRSTVWYid': 'RT348ACDEFGHKLMPQRSTVWYid',
                        '348I': 'RT348I',
                        '40F': 'RT40F',
                        '41L': 'RT41L',
                        '44AD': 'RT44AD',
                        '62V': 'RT62V',
                        '65ACDFGHILMPQSTVWYid': 'RT65ACDFGHILMPQSTVWYid',
                        '65E': 'RT65E',
                        '65N': 'RT65N',
                        '65R': 'RT65R',
                        '67ACFIKLMPQRVWYi': 'RT67ACFIKLMPQRVWYi',
                        '67GESTH': 'RT67GESTH',
                        '67N': 'RT67N',
                        '67d OR 68d OR 69d OR 70d': 'RT67-70d',
                        '68G': 'RT68G',
                        '69CFHKLMPQRVWY': 'RT69CFHKLMPQRVWY',
                        '69D': 'RT69D',
                        '69G': 'RT69G',
                        '69NSAIE': 'RT69NSAIE',
                        '69i': 'RT69i',
                        '70ACDFHILMPVWYi': 'RT70ACDFHILMPVWYi',
                        '70EGQNST': 'RT70EGQNST',
                        '70R': 'RT70R',
                        '74ACDEFGHKMNPQRSTWYid': 'RT74ACDEFGHKMNPQRSTWYid',
                        '74VI': 'RT74VI',
                        '75CDEFGHKNPQRWYid': 'RT75CDEFGHKNPQRWYid',
                        '75I': 'RT75I',
                        '75MTAS': 'RT75MTAS',
                        '77ACDEGHIKMNPQRSTVWYid': 'RT77ACDEGHIKMNPQRSTVWYid',
                        '77L': 'RT77L',
                        '90I': 'RT90I',
                        '98G': 'RT98G'}}

        res_comments = self.algorithm.parse_comments(self.root)

        self.assertEqual(exp_comments, res_comments)


if __name__ == '__main__':
    unittest.main()
